program Trader;   { Star Trader v4.1 By J.N.Zaitseff, 1991. }

{$U-}
{$C-}

const
  Version             = '4.1';
  MaxX                = 39;
  MaxY                = 15;
  MaxPlayers          = 8;
  MaxCompanies        = 8;
  DefaultMaxTurn      = 50;
  InitialReturn       = 0.05; { 5% }
  InitialCash         = 6000.0;
  NumberMoves         = 20;
  InitialInterestRate = 0.10; { 10% }

  Tab = ^I;
  Esc = ^[;

type
  str_       = string[40];
  CompanyRec = record
                 Name         : str_;
                 SharePrice   : real;
                 Return       : real;
                 StockIssued  : integer;
                 MaxStock     : integer;
                 OnMap        : boolean
               end;
  PlayerRec  = record
                 Name         : str_;
                 Cash         : real;
                 Debt         : real;
                 StockOwned   : array [1..MaxCompanies] of integer;
                 InGame       : boolean
               end;
  MapVal     = (Empty,Outpost,Star,C_A,C_B,C_C,C_D,C_E,C_F,C_G,C_H);
  MoveRec    = record
                 X, Y : integer
               end;

const
  CompanyVal : set of MapVal = [C_A..C_H];

var
  Company       : array [1..MaxCompanies] of CompanyRec;
  Player        : array [1..MaxPlayers] of PlayerRec;
  Map           : array [1..MaxX,1..MaxY] of MapVal;
  Move          : array [1..NumberMoves] of MoveRec;
  Selection     : char;
  InterestRate  : real;
  QuitSelected  : boolean;
  TurnNumber    : integer;
  MaxTurn       : integer;
  CurrentPlayer : integer;
  NumberPlayers : integer;
  FirstPlayer   : integer;
  GameLoaded    : boolean;
  GameNumber    : integer;

function rnd (N : real) : real;

var
  I : integer;
  R : real;

begin
  I := random(MaxInt);
  R := I/MaxInt;
  rnd := R + N
end;

function ltab (N : integer) : char;

var
  I : integer;

begin
  write(#13);
  for I := 1 to N do
    write(#12);
  ltab := #8
end;

procedure ClrEOS;

begin
  write(#27#89)
end;

procedure TextDump;

var
  X, Y, I : byte;

begin
  writeln(lst,#27'R'#0#27'P');
  for Y := 0 to 23 do
    begin
      for X := 0 to 79 do
        begin
          I := mem[$F000+X+Y*80];
          if I < 128 then
            write(lst,chr(I))
          else
            write(lst,#27'E',chr(I),#27'F')
        end;
      writeln(lst)
    end;
  writeln(lst)
end;

function Key : char;

var
  C : char;

begin
  repeat
    read(kbd,C);
    if C = ^P then
      TextDump
  until C <> ^P;
  Key := C
end;

function UKey : char;

begin
  UKey := upcase(Key)
end;

function Answer : char;

var
  C : char;

begin
  repeat
    C := UKey
  until C in ['Y','N'];
  Answer := C;
  if C = 'Y' then
    writeln('Yes.')
  else
    writeln('No.')
end;

function InputString : str_;

var
  S : str_;
  C : char;
  I : integer;

begin
  S := '';
  repeat
    C := Key;
    case C of
      ^H, #127 : if S <> '' then
                   begin
                     delete(S,length(S),1);
                     write(^H' '^H)
                   end;
      ^X       : if S <> '' then
                   begin
                     for I := 1 to length(S) do
                       write(^H' '^H);
                     S := ''
                   end;
      ^M, ^J   : C := ^M
    else
      if (C in ['A'..'Z','a'..'z','0'..'9',' ']) and (length(S) < 40) then
        begin
          S := S + C;
          write(C)
        end
    end
  until C = ^M;
  InputString := S
end;

function InputNumber : real;

var
  Temp : real;
  S    : str_;
  C    : char;
  I    : integer;

begin
  S := '';
  repeat
    C := Key;
    case C of
      ^H, #127 : if S <> '' then
                   begin
                     delete(S,length(S),1);
                     write(^H' '^H)
                   end;
      ^X       : if S <> '' then
                   begin
                     for I := 1 to length(S) do
                       write(^H' '^H);
                     S := ''
                   end;
      ^M, ^J   : C := ^M
    else
      if (C in ['0'..'9','.']) and (length(S) < 18) then
        begin
          S := S + C;
          write(C)
        end
    end
  until C = ^M;
  if S = '' then
    S := '0';
  val(S,Temp,I);
  InputNumber := Temp
end;

procedure Title;

begin
  ClrScr;
  GotoXY(28,1);
  LowVideo;
  writeln(' ***  Star Traders  *** ');
  NormVideo;
  GotoXY(29,2);
  writeln('By J.N.Zaitseff, 1991.');
  writeln
end;

procedure Continue;

var
  C : char;

begin
  write(Tab,Tab,Tab,'  [ Press any key to continue ] ');
  C := Key;
  DelLine
end;

procedure InitData;

var
  X, Y : integer;

begin
  QuitSelected := false;
  randomize;
  if not GameLoaded then
    begin
      Company[1].Name := 'Altair Starways';
      Company[2].Name := 'Betelgeuse, Ltd';
      Company[3].Name := 'Capella Freight Co';
      Company[4].Name := 'Denebola Shippers';
      Company[5].Name := 'Eridani Expediters';
      Company[6].Name := 'Fornax Express';
      Company[7].Name := 'Gemeni Inc';
      Company[8].Name := 'Hercules & Co';
      for X := 1 to MaxCompanies do
        begin
          Company[X].SharePrice := 0.0;
          Company[X].Return := InitialReturn;
          Company[X].StockIssued := 0;
          Company[X].MaxStock := 0;
          Company[X].OnMap := false
        end;
      InterestRate := InitialInterestRate;
      for X := 1 to MaxX do
        for Y := 1 to MaxY do
          if random(20) > 17 then
            Map[X,Y] := Star
          else
            Map[X,Y] := Empty
    end
end;

{$I-}

procedure Instructions;

var
  C        : char;
  F        : text;
  Inverted : boolean;
  B        : byte;

begin
  Title;
  assign(F,'TRADER.HLP');
  reset(F);
  if IOResult <> 0 then
    begin
      writeln;
      writeln('Help not available: overlay file TRADER.HLP not on disk.');
      writeln;
      Continue
    end
  else
    begin
      Inverted := false;
      while not eof(F) do
        begin
          read(F,C);
          B := ord(C);
          if B > $7F then
            C := chr(B-$80);
          case C of
            ^B  : begin
                    Inverted := not Inverted;
                    if Inverted then
                      LowVideo
                    else
                      NormVideo
                  end;
            ^A  : Continue;
            ^F  : begin
                    Continue;
                    Title
                  end;
            '~' : begin
                    read(F,C);
                    case C of
                      '~' : write(C);
                      'x' : write(MaxX:2);
                      'y' : write(MaxY:2);
                      'm' : write(NumberMoves:2);
                      'M' : write(chr(NumberMoves+$60));
                      'c' : write(MaxCompanies:2);
                      't' : write(MaxTurn:3)
                    end
                  end
          else
            write(C)
          end
        end;
      close(F);
      Continue
    end
end;

{$I+}

procedure InitPlayers;

var
  C    : char;
  I, J : integer;
  S    : str_;

{$I-}

procedure LoadGame (GN : integer);

var
  F : text;
  S : str_;
  K : integer;

begin
  S := 'A:TRADER'+chr(GN+$30)+'.GME';
  assign(F,S);
  reset(F);
  if IOResult <> 0 then
    begin
      writeln;
      writeln(tab,'That game has not been saved to disk.');
      writeln;
      Continue;
      C := ' ';
      exit
    end;
  readln(F,S);
  readln(F,S);
  if S <> Version then
    begin
      close(F);
      writeln;
      writeln(tab,'That game was saved under an old version of Star Traders.');
      writeln;
      Continue;
      C := ' ';
      exit
    end;
  readln(F,MaxTurn);
  readln(F,NumberPlayers);
  for I := 1 to NumberPlayers do
    begin
      readln(F,Player[I].Name);
      readln(F,Player[I].Cash);
      readln(F,Player[I].Debt);
      readln(F,J);
      if J = 0 then
        Player[I].InGame := true
      else
        Player[I].InGame := false;
      for J := 1 to MaxCompanies do
        readln(F,Player[I].StockOwned[J])
    end;
  for I := 1 to MaxCompanies do
    begin
      readln(F,Company[I].Name);
      readln(F,Company[I].SharePrice);
      readln(F,Company[I].Return);
      readln(F,Company[I].StockIssued);
      readln(F,Company[I].MaxStock);
      readln(F,J);
      if J = 0 then
        Company[I].OnMap := true
      else
        Company[I].OnMap := false
    end;
  readln(F,InterestRate);
  readln(F,TurnNumber);
  readln(F,CurrentPlayer);
  readln(F,FirstPlayer);
  for I := 1 to MaxX do
    for J := 1 to MaxY do
      begin
        read(F,C);
        Map[I,J] := MapVal(ord(C)-$40)
      end;
  close(F);
  C := 'C';
  QuitSelected := false;
  GameNumber := GN;
  GameLoaded := true
end;

{$I+}

begin { InitPlayers }
  GameLoaded := false;
  GameNumber := 0;
  if ParamCount = 2 then
    if (ParamStr(1) = 'GAME') and (ParamStr(2) >= '1') and (ParamStr(2) <= '9') then
      begin
        Title;
        writeln;
        write('Loading game ',ParamStr(2),'... ');
        C := ParamStr(2);
        LoadGame(ord(C)-$30);
        if GameLoaded then
          exit
      end;
  if ParamCount = 1 then
    begin
      I := pos('.',ParamStr(1));
      S := copy(ParamStr(1),I,4);
      if S = '.GME' then
        begin
          S := ParamStr(1);
          C := S[7];
          if C in ['1'..'9'] then
            begin
              Title;
              writeln;
              write('Loading game ',C,'... ');
              LoadGame(ord(C)-$30);
              if GameLoaded then
                exit
            end
        end
    end;
  repeat
    Title;
    NumberPlayers := 0;
    write('Enter number of players [1-',MaxPlayers,'] or <C> to continue a game: ');
    repeat
      C := UKey;
      NumberPlayers := ord(C)-$30
    until (NumberPlayers in [1..MaxPlayers]) or (C = 'C');
    writeln(C);
    if C = 'C' then
      begin
        writeln;
        write('Enter game number [1-9] or <ESC> to cancel: ');
        repeat
          C := Key
        until C in ['1'..'9',Esc];
        if C <> Esc then
          begin
            writeln(C);
            LoadGame(ord(C)-$30)
          end
      end
  until (C = 'C') or (NumberPlayers in [1..MaxPlayers]);
  if C <> 'C' then
    begin
      MaxTurn := DefaultMaxTurn;
      if ParamStr(1) = 'MAXTURN' then
        begin
          val(ParamStr(2),I,J);
          if (J = 0) and (I >= 10) and (I <= 500) then
            MaxTurn := I
        end;
      for I := 1 to NumberPlayers do
        begin
          writeln;
          write('Player #',I,'. Enter your name: ');
          Player[I].Name := '';
          repeat
            GotoXY(29,5+I);
            ClrEOL;
            Player[I].Name := InputString
          until Player[I].Name <> '';
          Player[I].Cash := InitialCash;
          Player[I].Debt := 0.0;
          Player[I].InGame := true;
          for J := 1 to MaxCompanies do
            Player[I].StockOwned[J] := 0
        end;
      writeln;
      writeln;
      if NumberPlayers = 1 then
        write('Do you need any instructions? [Y/N] ')
      else
        write('Does any player need instructions? [Y/N] ');
      C := Answer;
      if C = 'Y' then
        Instructions;
      Title;
      writeln;
      writeln(tab,tab,tab,'The game ends after ',MaxTurn,' moves.');
      writeln;
      Continue;
      Title;
      TurnNumber := 1;
      CurrentPlayer := 1;
      FirstPlayer := 1;
      if NumberPlayers <> 1 then
        begin
          FirstPlayer := random(NumberPlayers)+1;
          CurrentPlayer := FirstPlayer;
          writeln(Tab,'The first player to go is ',Player[FirstPlayer].Name,'.');
          writeln;
          Continue
        end
    end
end;

procedure NextPlayer;

var
  AllOut : boolean;
  I      : integer;

begin
  AllOut := true;
  for I := 1 to NumberPlayers do
    if Player[I].InGame then
      AllOut := false;
  if not AllOut then
    repeat
      CurrentPlayer := CurrentPlayer + 1;
      if CurrentPlayer = NumberPlayers + 1 then
        CurrentPlayer := 1;
      if CurrentPlayer = FirstPlayer then
        TurnNumber := TurnNumber + 1
    until Player[CurrentPlayer].InGame
end;

procedure SelectMoves;

procedure SortMoves;

var
  I, J, K : integer;
  Temp    : MoveRec;

begin
  for I := 1 to NumberMoves-1 do
    begin
      K := I;
      Temp := Move[I];
      for J := I+1 to NumberMoves do
        if Move[J].X < Temp.X then
          begin
            K := J;
            Temp := Move[J]
          end;
      Move[K] := Move[I];
      Move[I] := Temp
    end
end;

var
  I, J, TempX, TempY : integer;
  OK                 : boolean;

begin { SelectMoves }
  I := 0;
  for TempX := 1 to MaxX do
    for TempY := 1 to MaxY do
      if Map[TempX,TempY] = Empty then
        I := I + 1;
  if I < NumberMoves then
    QuitSelected := true
  else
    begin
      for I := 1 to NumberMoves do
        begin
          repeat
            TempX := random(MaxX)+1;
            TempY := random(MaxY)+1;
            OK := true;
            if I > 1 then
              for J := I-1 downto 1 do
                if ((Move[J].X = TempX) and (Move[J].Y = TempY)) then
                  OK := false
          until OK and (Map[TempX,TempY] = Empty);
          Move[I].X := TempX;
          Move[I].Y := TempY
        end;
      QuitSelected := false;
      SortMoves
    end
end;

procedure DisplayMap;

var
  X, Y, I : integer;

begin
  Title;
  write('Current Player: ',Player[CurrentPlayer].Name,'. Turn #',TurnNumber);
  if TurnNumber = MaxTurn then
    write(' *** Last turn ***');
  writeln;
  writeln;
  for Y := 1 to MaxY do
    begin
      for X := 1 to MaxX do
        case Map[X,Y] of
          Empty   : write(' .');
          Outpost : write(' +');
          Star    : write(' *')
        else
          write(' ',chr(ord(Map[X,Y])-ord(C_A)+65))
        end;
      writeln
    end
end;

function TotalValue (N : integer) : real;

var
  I    : integer;
  Temp : real;

begin
  Temp := Player[N].Cash - Player[N].Debt;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      Temp := Temp + Player[N].StockOwned[I]*Company[I].SharePrice;
  TotalValue := Temp
end;

procedure DisplayStatus;

var
  Temp      : real;
  I         : integer;
  AllOffMap : boolean;

begin
  Title;
  writeln('Player: ',Player[CurrentPlayer].Name,'.',tab,'Stock owned.');
  writeln;
  if TotalValue(CurrentPlayer) = 0 then
    writeln('***  Bankrupt  ***')
  else
    begin
      AllOffMap := true;
      for I := 1 to MaxCompanies do
        if Company[I].OnMap then
          AllOffMap := false;
      if AllOffMap then
        writeln('No companies on the map.')
      else
        begin
          writeln(tab,tab,tab,'Price per',tab,tab,'Holdings   Ownership of');
          write('Company  ',tab,tab,'share ($)   Return (%)  (shares)');
          writeln('   company (%)');
          writeln;
          for I := 1 to MaxCompanies do
            if Company[I].OnMap then
              begin
                write(Company[I].Name,ltab(25));
                write(Company[I].SharePrice:9:2,'   ');
                write(Company[I].Return*100:10:2,'  ');
                write(Player[CurrentPlayer].StockOwned[I]:8,'   ');
                if Company[I].StockIssued = 0 then
                  Temp := 0.0
                else
                  Temp := Player[CurrentPlayer].StockOwned[I]
                          / Company[I].StockIssued;
                writeln(Temp*100:12:2)
              end
        end;
      writeln;
      writeln('Current cash: $',Player[CurrentPlayer].Cash:4:2);
      if Player[CurrentPlayer].Debt <> 0 then
        begin
          writeln('Current debt: $',Player[CurrentPlayer].Debt:4:2);
          writeln('Debt interest rate: ',InterestRate*100:4:2,'%')
        end;
      writeln;
      writeln('Total value: $',TotalValue(CurrentPlayer):4:2)
    end;
  writeln;
  Continue
end;

procedure GetMove;

var
  I, X, Y : integer;
  C       : char;

begin
  if QuitSelected then
    Selection := Esc
  else
    begin
      repeat
        DisplayMap;
        for I := 1 to NumberMoves do
          begin
            X := Move[I].X;
            Y := Move[I].Y;
            GotoXY(X*2,Y+5);
            LowVideo;
            write(chr(I+96));
            NormVideo
          end;
        repeat
          GotoXY(1,22);
          ClrEOS;
          writeln;
          LowVideo;
          write('<1>');
          NormVideo;
          write(' Display stock owned.  ');
          LowVideo;
          write('<2>');
          NormVideo;
          writeln(' Declare bankruptcy.  ');
          LowVideo;
          write('<3>');
          NormVideo;
          write(' Save game.            ');
          LowVideo;
          write('<ESC>');
          NormVideo;
          write(' Quit the game.');
          GotoXY(1,22);
          write('Select move [a-',chr(NumberMoves+96),'/1/2/3/<ESC>]: ');
          repeat
            C := UKey;
            if C in ['1','2','3',Esc] then
              I := 1
            else
              I := ord(C)-$40
          until I in [1..NumberMoves];
          if C = Esc then
            writeln('<ESC>')
          else
            writeln(C);
          ClrEOS;
          Selection := C;
          if Selection = '1' then
            DisplayStatus
          else
            begin
              write('Are you sure? [Y/N] ');
              C := Answer
            end
        until C in ['Y','1']
      until C <> '1'
    end
end;

procedure ProcessMove;

var
  X, Y, I                     : integer;
  Left, Right, Up, Down, Temp : MapVal;
  AllOut                      : boolean;

procedure AssignSurroundVals;

begin
  if X = 1    then Left  := Empty else Left  := Map[X-1,Y];
  if X = MaxX then Right := Empty else Right := Map[X+1,Y];
  if Y = 1    then Up    := Empty else Up    := Map[X,Y-1];
  if Y = MaxY then Down  := Empty else Down  := Map[X,Y+1]
end;

procedure IncreaseSharePrice (N : integer; P : real);

var
  I : integer;

begin
  I := round(P);
  Company[N].SharePrice := Company[N].SharePrice + P + (P * rnd(0) / 2);
  Company[N].MaxStock := Company[N].MaxStock + (I div (random(10)+5));
  if rnd(0) < 0.25 then
    Company[N].Return := Company[N].Return * rnd(0.60)
end;

{$A-}

procedure IncludeOutpost (Xn, Yn : integer);

var
  L, R, U, D : MapVal;

begin
  Map[Xn,Yn] := Temp;
  IncreaseSharePrice(I,70.00);
  if Xn = 1    then L := Empty else L := Map[Xn-1,Yn];
  if Xn = MaxX then R := Empty else R := Map[Xn+1,Yn];
  if Yn = 1    then U := Empty else U := Map[Xn,Yn-1];
  if Yn = MaxY then D := Empty else D := Map[Xn,Yn+1];
  if L = Outpost then IncludeOutpost(Xn-1,Yn);
  if R = Outpost then IncludeOutpost(Xn+1,Yn);
  if U = Outpost then IncludeOutpost(Xn,Yn-1);
  if D = Outpost then IncludeOutpost(Xn,Yn+1)
end;

{$A+}

procedure EstablishNewCompany;

var
  I, J    : integer;
  AllUsed : boolean;
  Temp    : MapVal;

const
  CompanyFormers : set of MapVal = [Outpost,Star];

begin
  if (Left in CompanyFormers) or (Right in CompanyFormers) or
     (Up   in CompanyFormers) or (Down  in CompanyFormers) then
    begin
      AllUsed := true;
      for I := MaxCompanies downto 1 do
        if not Company[I].OnMap then
          begin
            AllUsed := false;
            J := I
          end;
      if AllUsed then
        Map[X,Y] := Outpost
      else
        begin
          Title;
          writeln;
          write('A new company has been formed.');
          writeln(' Its name is ',Company[J].Name,'.');
          writeln;
          writeln;
          Continue;
          Temp := MapVal(J+ord(C_A)-1);
          Map[X,Y] := Temp;
          Company[J].Return := InitialReturn;
          Company[J].SharePrice := 60.00;
          Company[J].OnMap := true;
          Company[J].StockIssued := 5;
          Company[J].MaxStock := 50;
          for I := 1 to NumberPlayers do
            Player[I].StockOwned[J] := 0;
          Player[CurrentPlayer].StockOwned[J] := 5
        end
    end
end;

procedure MergeCompanies (Company1,Company2 : MapVal);

var
  I, J, X1, Y1      : integer;
  Bonus, Val1, Val2 : real;
  ExtraStock        : integer;
  SwapTemp          : MapVal;

begin
  I := ord(Company1)-ord(C_A)+1;
  J := ord(Company2)-ord(C_A)+1;
  Val1 := Company[I].SharePrice * Company[I].StockIssued * Company[I].Return;
  Val2 := Company[J].SharePrice * Company[J].StockIssued * Company[J].Return;
  if Val1 < Val2 then
    begin
      SwapTemp := Company1;
      Company1 := Company2;
      Company2 := SwapTemp;
      X1 := I;
      I := J;
      J := X1
    end;
  Title;
  writeln(Company[J].Name,' has just merged into ',Company[I].Name,'.');
  writeln;
  writeln('Please note the following transactions:');
  writeln;
  writeln('Old stock: ',Company[J].Name,'.');
  writeln('New stock: ',Company[I].Name,'.');
  writeln;
  write('Player',ltab(25),'Old stock   New stock   Total holdings');
  writeln('  Bonus paid ($)');
  writeln;
  Y1 := 0;
  for X1 := 1 to NumberPlayers do
    if Player[X1].InGame then
      begin
        write(Player[X1].Name,ltab(25));
        ExtraStock := Player[X1].StockOwned[J] div 2;
        write(Player[X1].StockOwned[J]:9,'   ',ExtraStock:9,'   ');
        Player[X1].StockOwned[I] := Player[X1].StockOwned[I] + ExtraStock;
        Y1 := Y1 + ExtraStock;
        write(Player[X1].StockOwned[I]:14,'  ');
        if Company[J].StockIssued = 0 then
          Bonus := 0.0
        else
          Bonus := 10 * ((Player[X1].StockOwned[J]/Company[J].StockIssued)
                   * Company[J].SharePrice);
        writeln(Bonus:14:2);
        Player[X1].Cash := Player[X1].Cash + Bonus;
        Player[X1].StockOwned[J] := 0
      end;
  writeln;
  Continue;
  Map[X,Y] := Company2;
  Company[I].StockIssued := Company[I].StockIssued + Y1;
  Company[I].SharePrice := Company[I].SharePrice + Company[J].SharePrice / rnd(1.5);
  Company[I].MaxStock := Company[I].MaxStock + Y1;
  Company[J].StockIssued := 0;
  Company[J].MaxStock := 0;
  Company[J].OnMap := false;
  for X1 := 1 to MaxX do
    for Y1 := 1 to MaxY do
      if Map[X1,Y1] = Company2 then
        Map[X1,Y1] := Company1;
  AssignSurroundVals
end;

procedure SaveGame;

var
  F : text;
  C : char;
  S : str_;

begin
  Title;
  if GameLoaded then
    C := chr(GameNumber+$30)
  else
    begin
      writeln;
      writeln;
      write('Enter game number to save [1-9]: ');
      repeat
        C := Key
      until C in ['1'..'9'];
      writeln(C)
    end;
  writeln;
  writeln;
  write('Saving game... ');
  S := 'A:TRADER'+C+'.GME';
  assign(F,S);
  rewrite(F);
  writeln(F,'@Trader');
  writeln(F,Version);
  writeln(F,MaxTurn);
  writeln(F,NumberPlayers);
  for I := 1 to NumberPlayers do
    begin
      writeln(F,Player[I].Name);
      writeln(F,Player[I].Cash);
      writeln(F,Player[I].Debt);
      if Player[I].InGame then
        writeln(F,'0')
      else
        writeln(F,'1');
      for X := 1 to MaxCompanies do
        writeln(F,Player[I].StockOwned[X])
    end;
  for I := 1 to MaxCompanies do
    begin
      writeln(F,Company[I].Name);
      writeln(F,Company[I].SharePrice);
      writeln(F,Company[I].Return);
      writeln(F,Company[I].StockIssued);
      writeln(F,Company[I].MaxStock);
      if Company[I].OnMap then
        writeln(F,'0')
      else
        writeln(F,'1')
    end;
  writeln(F,InterestRate);
  writeln(F,TurnNumber);
  writeln(F,CurrentPlayer);
  writeln(F,FirstPlayer);
  for X := 1 to MaxX do
    for Y := 1 to MaxY do
      write(F,chr(ord(Map[X,Y])+$40));
  writeln(F);
  writeln(F,'#Trader');
  close(F);
  Selection := Esc
end;

begin { ProcessMove }
  Title;
  if Selection = '2' then
    begin
      writeln(Player[CurrentPlayer].Name,' has declared bankruptcy.');
      writeln;
      Continue;
      DisplayStatus;
      Title;
      Player[CurrentPlayer].InGame := false;
      for I := 1 to MaxCompanies do
        begin
          Company[I].MaxStock := Company[I].MaxStock +
                                 Player[CurrentPlayer].StockOwned[I];
          Company[I].StockIssued := Company[I].StockIssued -
                                    Player[CurrentPlayer].StockOwned[I];
          Player[CurrentPlayer].StockOwned[I] := 0
        end;
      Player[CurrentPlayer].Cash := 0;
      Player[CurrentPlayer].Debt := 0;
      AllOut := true;
      for I := 1 to NumberPlayers do
        if Player[I].InGame then
          AllOut := false;
      if AllOut then
        Selection := Esc
    end;
  if Selection = '3' then
    SaveGame;
  if Selection = Esc then
    begin
      QuitSelected := true;
      exit
    end;
  if Player[CurrentPlayer].InGame then
    begin
      I := ord(Selection) - $40;
      X := Move[I].X;
      Y := Move[I].Y;
      AssignSurroundVals;
      if (Left = Empty) and (Right = Empty) and (Up = Empty) and (Down = Empty) then
        Map[X,Y] := Outpost;
      if (Left in CompanyVal) and (Right in CompanyVal) and (Left <> Right) then
        MergeCompanies(Left,Right);
      if (Left in CompanyVal) and (Up in CompanyVal) and (Left <> Up) then
        MergeCompanies(Left,Up);
      if (Left in CompanyVal) and (Down in CompanyVal) and (Left <> Down) then
        MergeCompanies(Left,Down);
      if (Right in CompanyVal) and (Up in CompanyVal) and (Right <> Up) then
        MergeCompanies(Right,Up);
      if (Right in CompanyVal) and (Down in CompanyVal) and (Right <> Down) then
        MergeCompanies(Right,Down);
      if (Up in CompanyVal) and (Down in CompanyVal) and (Up <> Down) then
        MergeCompanies(Up,Down);
      if not((Left in CompanyVal) or (Right in CompanyVal) or (Up in CompanyVal)
         or (Down in CompanyVal)) then
        EstablishNewCompany;
      Temp := Empty;
      if Left  in CompanyVal then Temp := Left;
      if Right in CompanyVal then Temp := Right;
      if Up    in CompanyVal then Temp := Up;
      if Down  in CompanyVal then Temp := Down;
      if Temp <> Empty then
        begin
          I := ord(Temp)-ord(C_A)+1;
          Map[X,Y] := Temp;
          IncreaseSharePrice(I,60.0)
        end;
      Temp := Map[X,Y];
      I := 0;
      if Temp in CompanyVal then I := ord(Temp)-ord(C_A)+1;
      if I > 0 then
        begin
          if Left  = Star then IncreaseSharePrice(I,300.0);
          if Right = Star then IncreaseSharePrice(I,300.0);
          if Up    = Star then IncreaseSharePrice(I,300.0);
          if Down  = Star then IncreaseSharePrice(I,300.0)
        end;
      if Temp in CompanyVal then
        begin
          if Left  = Outpost then IncludeOutpost(X-1,Y);
          if Right = Outpost then IncludeOutpost(X+1,Y);
          if Up    = Outpost then IncludeOutpost(X,Y-1);
          if Down  = Outpost then IncludeOutpost(X,Y+1)
        end
    end
end;

procedure StockExchange;

var
  AllOff  : boolean;
  I, J    : integer;
  C       : char;
  BidUsed : boolean;
  R, R2   : real;

procedure AdjustValues;

var
  X, Y : integer;
  Temp : MapVal;

begin
  if rnd(0) > 0.99 then
    begin
      I := random(MaxCompanies)+1;
      if Company[I].OnMap then
        begin
          Title;
          writeln;
          write(Company[I].Name,' has been declared bankrupt ');
          writeln('by the Interstellar Trading Bank.');
          writeln;
          R := 0;
          if rnd(0) > 0.2 then
            R := rnd(0);
          if R = 0 then
            writeln('All assets have been taken to repay outstanding loans.')
          else
            begin
              write('The bank has agreed to pay stock holders ',R*100:4:2);
              writeln('% of the share value on each share owned.');
              writeln;
              writeln('Old share value:       $',Company[I].SharePrice:4:2);
              writeln('Amount paid per share: $',Company[I].SharePrice * R:4:2);
              writeln;
              writeln('Player',ltab(25),'Holdings   Value ($)   Amount paid ($)');
              writeln;
              for J := 1 to NumberPlayers do
                if Player[J].InGame then
                  begin
                    write(Player[J].Name,ltab(25),Player[J].StockOwned[I]:8);
                    R2 := Player[J].StockOwned[I] * Company[I].SharePrice;
                    write('   ',R2:9:2);
                    writeln('   ',R2 * R:15:2);
                    Player[J].Cash := Player[J].Cash + R2 * R
                  end
            end;
          writeln;
          Continue;
          Company[I].OnMap := false;
          for J := 1 to NumberPlayers do
            Player[J].StockOwned[I] := 0;
          Company[I].SharePrice := 0.0;
          Company[I].Return := 0.0;
          Company[I].StockIssued := 0;
          Company[I].MaxStock := 0;
          Temp := MapVal(I+ord(C_A)-1);
          for X := 1 to MaxX do
            for Y := 1 to MaxY do
              if Map[X,Y] = Temp then
                Map[X,Y] := Empty
        end
    end;
  if rnd(0) < 0.4 then
    begin
      I := random(MaxCompanies)+1;
      Company[I].Return := Company[I].Return * rnd(0.75)
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap and (Company[I].Return > 0.40) then
      Company[I].Return := Company[I].Return / rnd(1.5);
  if rnd(0) < 0.3 then
    begin
      I := random(MaxCompanies)+1;
      R := rnd(0)*(Company[I].SharePrice/4);
      if rnd(0) < 0.65 then
        Company[I].SharePrice := Company[I].SharePrice - R
      else
        Company[I].SharePrice := Company[I].SharePrice + R
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      begin
        if Company[I].StockIssued = 0 then
          R := 0
        else
          R := Player[CurrentPlayer].StockOwned[I]/Company[I].StockIssued;
        Player[CurrentPlayer].Cash := Player[CurrentPlayer].Cash +
          (Player[CurrentPlayer].StockOwned[I] * Company[I].SharePrice *
          Company[I].Return) + (R * Company[I].SharePrice * 2)
      end;
  if rnd(0) < 0.3 then
    InterestRate := InterestRate * rnd(0.65);
  if InterestRate > 0.30 then
    InterestRate := InterestRate / rnd(1.5);
  if (Player[CurrentPlayer].Debt * InterestRate) > 10000.00 then
    R := 0.5
  else
    R := 1;
  Player[CurrentPlayer].Debt := Player[CurrentPlayer].Debt * (1 + InterestRate * R)
end;

procedure Title2;

begin
  Title;
  GotoXY(22,4);
  writeln('***  Interstellar Stock Exchange  ***');
  writeln('Player: ',Player[CurrentPlayer].Name)
end;

procedure VisitBank;

var
  C     : char;
  R, R2 : real;

begin
  Title2;
  writeln(ltab(24),'=== Interstellar Trading Bank ===');
  writeln;
  writeln('Current cash: $',Player[CurrentPlayer].Cash:4:2);
  writeln('Current debt: $',Player[CurrentPlayer].Debt:4:2);
  writeln('Interest rate: ',InterestRate*100:4:2,'%');
  R2 := 2 * (TotalValue(CurrentPlayer) - Player[CurrentPlayer].Debt);
  if R2 < 0 then
    R2 := 0;
  writeln('Credit limit: $',R2:4:2);
  writeln;
  LowVideo;
  write('<1>');
  NormVideo;
  writeln(' Borrow money.');
  LowVideo;
  write('<2>');
  NormVideo;
  writeln(' Repay debt.');
  LowVideo;
  write('<3>');
  NormVideo;
  writeln(' Exit from Bank.');
  writeln;
  write('Enter selection [1/2/3]: ');
  repeat
    C := Key
  until C in ['1'..'3',' '];
  writeln(C);
  writeln;
  case C of
    '1' : if R2 = 0 then
            begin
              writeln('The bank will not lend you any more money.');
              writeln;
              Continue
            end
          else
            begin
              write('How much do you wish to borrow? ');
              repeat
                GotoXY(33,19);
                ClrEOL;
                R := InputNumber
              until R <= (R2 + 0.01);
              Player[CurrentPlayer].Cash := Player[CurrentPlayer].Cash + R;
              Player[CurrentPlayer].Debt := (1+InterestRate) *
                                            (Player[CurrentPlayer].Debt + R)
            end;
    '2' : begin
            if Player[CurrentPlayer].Debt = 0 then
              begin
                writeln('You have no debt to pay off.');
                writeln;
                Continue
              end
            else
              begin
                write('How much do you wish to repay? ');
                repeat
                  GotoXY(32,19);
                  ClrEOL;
                  R := InputNumber
                until (R <= (Player[CurrentPlayer].Cash+0.01)) and
                      (R <= (Player[CurrentPlayer].Debt+0.01));
                Player[CurrentPlayer].Cash := Player[CurrentPlayer].Cash - R;
                Player[CurrentPlayer].Debt := Player[CurrentPlayer].Debt - R;
                if Player[CurrentPlayer].Debt < 0.01 then
                  Player[CurrentPlayer].Debt := 0;
                if Player[CurrentPlayer].Cash < 0.01 then
                  Player[CurrentPlayer].Cash := 0
              end
          end
  end
end;

procedure Trade (N : integer);

var
  R, R2 : real;
  C     : char;
  I, J  : integer;

procedure Title3;

begin
  Title2;
  writeln;
  writeln('--- Stock transaction in ',Company[N].Name,' ---');
  writeln;
  writeln('Current cash:     $',Player[CurrentPlayer].Cash:4:2);
  writeln('Current holdings: ',Player[CurrentPlayer].StockOwned[N]);
  writeln;
  writeln('Each share costs $',Company[N].SharePrice:4:2);
  writeln('There are ',Company[N].MaxStock-Company[N].StockIssued,' shares left.')
end;

begin
  Title2;
  writeln;
  writeln('Company name:  ',Company[N].Name);
  writeln('Shares issued: ',Company[N].StockIssued);
  writeln('Shares left:   ',Company[N].MaxStock-Company[N].StockIssued);
  writeln('Price:         $',Company[N].SharePrice:4:2);
  writeln('Return:        ',Company[N].Return*100:4:2,'%');
  writeln;
  writeln('Current cash:     $',Player[CurrentPlayer].Cash:4:2);
  writeln('Current holdings: ',Player[CurrentPlayer].StockOwned[N]);
  if Company[N].StockIssued = 0 then
    R2 := 0.0
  else
    R2 := Player[CurrentPlayer].StockOwned[N]/Company[N].StockIssued;
  writeln('Percentage owned: ',R2*100:4:2,'%');
  writeln;
  LowVideo;
  write('<1>');
  NormVideo;
  write(' - Buy stock.',tab,tab,tab,tab);
  LowVideo;
  write('<2>');
  NormVideo;
  writeln(' - Sell stock.');
  LowVideo;
  write('<3>');
  NormVideo;
  write(' - Bid company to issue more shares.',tab,tab);
  LowVideo;
  write('<4>');
  NormVideo;
  writeln(' - Exit.');
  writeln;
  write('Enter selection [1-4]: ');
  repeat
    C := Key
  until C in ['1'..'4',' '];
  writeln(C);
  case C of
    '1' : begin
            Title3;
            R := Player[CurrentPlayer].Cash/Company[N].SharePrice;
            if R > 32000 then
              R := 32000;
            I := trunc(R);
            if I > (Company[N].MaxStock - Company[N].StockIssued) then
              I := Company[N].MaxStock - Company[N].StockIssued;
            writeln('You can purchase ',I,' shares.');
            writeln;
            write('How many shares do you wish to purchase? ');
            repeat
              GotoXY(42,16);
              ClrEOL;
              R := InputNumber;
              if R > 32000 then
                R := 32000;
              J := trunc(R)
            until (J >= 0) and (J <= I);
            Player[CurrentPlayer].StockOwned[N] := Player[CurrentPlayer].
                                                   StockOwned[N] + J;
            Company[N].StockIssued := Company[N].StockIssued + J;
            Player[CurrentPlayer].Cash := Player[CurrentPlayer].Cash -
                                          (J * Company[N].SharePrice)
          end;
    '2' : if Player[CurrentPlayer].StockOwned[N] = 0 then
            begin
              Title2;
              writeln;
              writeln('You do not have any shares to sell.');
              writeln;
              Continue
            end
          else
            begin
              Title3;
              writeln;
              write('How many shares do you wish to sell? ');
              repeat
                GotoXY(38,15);
                ClrEOL;
                R := InputNumber;
                if R > 32000 then
                  R := 32000;
                J := trunc(R)
              until (J >= 0) and (J <= Player[CurrentPlayer].StockOwned[N]);
              Player[CurrentPlayer].StockOwned[N] := Player[CurrentPlayer].
                                                     StockOwned[N] - J;
              Company[N].StockIssued := Company[N].StockIssued - J;
              Player[CurrentPlayer].Cash := Player[CurrentPlayer].Cash +
                                            (J * Company[N].SharePrice)
            end;
    '3' : begin
            R := 0;
            if (not BidUsed) and (rnd(0) < R2) and (rnd(0) < 0.75) then
              begin
                R := rnd(0) * R2 * 200;
                Company[N].MaxStock := Company[N].MaxStock + trunc(R)
              end;
            BidUsed := true;
            Title3;
            writeln;
            if trunc(R) = 0 then
              writeln(Company[N].Name,' has refused to issue more shares.')
            else
              writeln(Company[N].Name,' has issued ',trunc(R),' more shares.');
            writeln;
            Continue
          end
  end
end;

begin { StockExchange }
  if (not QuitSelected) and Player[CurrentPlayer].InGame then
    begin
      BidUsed := false;
      AdjustValues;
      if TotalValue(CurrentPlayer) <= -100000.0 then
        begin
          Title;
          writeln;
          writeln('Your debt has amounted to $',Player[CurrentPlayer].Debt:4:2);
          writeln;
          write('The bank has impounded $',Player[CurrentPlayer].Cash:4:2);
          write(' from your cash.');
          writeln;
          Continue;
          Player[CurrentPlayer].Debt := Player[CurrentPlayer].Debt / (InterestRate
                                        + 1) - Player[CurrentPlayer].Cash;
          Player[CurrentPlayer].Cash := 0;
          if (TotalValue(CurrentPlayer) <= 0) and (rnd(0) < 0.07) then
            begin
              Title;
              writeln;
              write('You have been declared bankrupt');
              writeln(' by the Interstellar Trading Bank.');
              writeln;
              Continue;
              DisplayStatus;
              Title;
              Player[CurrentPlayer].InGame := false;
              for I := 1 to MaxCompanies do
                begin
                  Company[I].MaxStock := Company[I].MaxStock +
                                         Player[CurrentPlayer].StockOwned[I];
                  Company[I].StockIssued := Company[I].StockIssued -
                                            Player[CurrentPlayer].StockOwned[I];
                  Player[CurrentPlayer].StockOwned[I] := 0
                end;
              Player[CurrentPlayer].Cash := 0;
              Player[CurrentPlayer].Debt := 0;
              AllOff := true;
              for I := 1 to NumberPlayers do
                if Player[I].InGame then
                  AllOff := false;
              if AllOff then
                QuitSelected := true;
              exit
            end
        end
    end;
  repeat
    Title2;
    writeln;
    AllOff := true;
    for I := 1 to MaxCompanies do
      if Company[I].OnMap then
        AllOff := false;
    if AllOff then
      writeln('No companies on the map.')
    else
      begin
        write('  Company',tab,tab,'Shares issued   Shares left   ');
        writeln('Price ($)   Return (%)');
        writeln;
        for I := 1 to MaxCompanies do
          if Company[I].OnMap then
            begin
              LowVideo;
              write(Company[I].Name[1]);
              NormVideo;
              write(' ',Company[I].Name,ltab(25));
              write(Company[I].StockIssued:13,'   ');
              write(Company[I].MaxStock-Company[I].StockIssued:11,'   ');
              write(Company[I].SharePrice:9:2,'   ');
              writeln(Company[I].Return*100:10:2)
            end
      end;
    writeln;
    writeln;
    LowVideo;
    write('<1>');
    NormVideo;
    write(' Display stock owned.',tab,tab,tab);
    LowVideo;
    write('<2>');
    NormVideo;
    writeln(' Display map.');
    LowVideo;
    write('<3>');
    NormVideo;
    write(' Visit the Interstellar Trading Bank.',tab);
    LowVideo;
    write('<4>');
    NormVideo;
    writeln(' Exit the Stock Exchange.');
    writeln;
    write('Enter selection [Company letter/1-4]: ');
    repeat
      C := UKey
    until (C in ['1'..'4',' ']) or ((C >= 'A') and (C <= chr(MaxCompanies+64))
          and Company[ord(C)-$40].OnMap);
    write(C);
    case C of
      '1'      : DisplayStatus;
      '2'      : begin
                   DisplayMap;
                   writeln;
                   Continue
                 end;
      '3'      : VisitBank;
      '4', ' ' : C := ' '
    else
      Trade(ord(C)-$40)
    end
  until C = ' '
end;

procedure DisplayStatus_of_all;

var
  I, J : integer;
  C    : char;

procedure SortPlayers;

var
  I, J, K : integer;
  Temp    : PlayerRec;
  R1, R2  : real;

begin
  for I := 1 to NumberPlayers-1 do
    begin
      K := I;
      Temp := Player[I];
      for J := I+1 to NumberPlayers do
        if TotalValue(K) < TotalValue(J) then
          begin
            K := J;
            Temp := Player[J]
          end;
      Player[K] := Player[I];
      Player[I] := Temp
    end
end;

begin
  Title;
  writeln;
  writeln;
  writeln('--- The game is over after ',TurnNumber-1,' turns ---');
  writeln;
  writeln;
  Continue;
  for I := 1 to NumberPlayers do
    begin
      CurrentPlayer := I;
      DisplayStatus
    end;
  Title;
  if NumberPlayers <> 1 then
    SortPlayers;
  if NumberPlayers = 1 then
    writeln('Your total value was $',TotalValue(1):4:2)
  else
    begin
      write('The winner is ',Player[1].Name);
      if TotalValue(1) = 0 then
        writeln(', who is *** Bankrupt ***.')
      else
        writeln(', with a value of $',TotalValue(1):4:2);
      writeln;
      writeln;
      writeln('Position',tab,'Player',ltab(40),'Total value ($)');
      writeln;

      for I := 1 to NumberPlayers do
        begin
          write(I);
          case I of
            1 : write('st');
            2 : write('nd');
            3 : write('rd')
          else
            write('th')
          end;
          write(tab,tab,Player[I].Name,ltab(40));
          if TotalValue(I) = 0 then
            writeln('*** BANKRUPT ***')
          else
            writeln(TotalValue(I):16:2)
        end
    end;
  writeln;
  writeln;
  write(tab,tab,tab,'[ Press <RETURN> to exit. ] ');
  repeat
    C := Key
  until C = ^M
end;


begin { Main }
  Title;
  InitPlayers;
  InitData;
  repeat
    SelectMoves;
    GetMove;
    ProcessMove;
    StockExchange;
    NextPlayer
  until (TurnNumber = MaxTurn + 1) or QuitSelected;
  DisplayStatus_of_all;
  ClrScr
end.
