program Trader;   { Star Trader By J.N.Zaitseff, 1991. }

{ Star Traders is a game of interstellar trading, where the object of the game
  is to create companies, buy and sell shares, borrow and repay money, in order
  to become the wealthiest player (the winner).

  Star Traders is executed thus:

TRADER <return> 			  - normal game.
TRADER n <return>			  - continue game "n" (1-9).

  This version of Star Traders is customised for the NEC Advanced Personal
  Computer running CP/M-86. All versions copyright by J.N.Zaitseff, 1992. }

{$U-,C-}

const
  Version	      = '4.4';
  MaxX		      = 39;
  MaxY		      = 15;
  MaxPlayers	      = 8;
  MaxCompanies	      = 8;
  DefaultMaxTurn      = 50;
  InitialReturn       = 0.05; { 5% }
  InitialCash	      = 6000.0;
  NumberMoves	      = 20;
  InitialInterestRate = 0.10; { 10% }
  StarRatio	      = 0.10; { 10% of map should be stars }
  LastParam	      = 10;

  Esc = ^[;

type
  str_	     = string[40];
  CompanyRec = record
		 Name	      : str_;
		 SharePrice   : real;
		 Return       : real;
		 StockIssued  : integer;
		 MaxStock     : integer;
		 OnMap	      : boolean
	       end;
  PlayerRec  = record
		 Name	      : str_;
		 Cash	      : real;
		 Debt	      : real;
		 StockOwned   : array [1..MaxCompanies] of integer;
		 InGame       : boolean
	       end;
  MapVal     = (Empty,Outpost,Star,C_A,C_B,C_C,C_D,C_E,C_F,C_G,C_H);
  MoveRec    = record
		 X, Y : integer
	       end;

const
  CompanyVal : set of MapVal = [C_A..C_H];

var
  Company	: array [1..MaxCompanies] of CompanyRec;
  Player	: array [1..MaxPlayers] of PlayerRec;
  Map		: array [1..MaxX,1..MaxY] of MapVal;
  Move		: array [1..NumberMoves] of MoveRec;
  Selection	: char;
  InterestRate	: real;
  QuitSelected	: boolean;
  TurnNumber	: integer;
  MaxTurn	: integer;
  CurrentPlayer : integer;
  NumberPlayers : integer;
  FirstPlayer	: integer;
  GameLoaded	: boolean;
  GameNumber	: integer;

  Parameter	: string[127] absolute DSeg:$0080;
  ParamStr	: array [1..LastParam] of str_;

{ Utility-type procedures and functions }

function ParamCount : integer;	   { How many parameters given on the parameter
				     line. Must be called before ParamStr is
				     examined. }

var
  I, J, Len : integer;

begin
  for I := 1 to LastParam do
    ParamStr[I] := '';
  Len := ord(Parameter[0]);
  I := 1;
  J := 1;
  while (I <= Len) and (J <= LastParam) do
    begin
      while (Parameter[I] = ' ') and (I <= Len) do
	I := I + 1;
      while (Parameter[I] <> ' ') and (I <= Len) do
	begin
	  ParamStr[J] := ParamStr[J] + Parameter[I];
	  I := I + 1
	end;
      while (Parameter[I] = ' ') and (I <= Len) do
	I := I + 1;
      J := J + 1
    end;
  ParamCount := J - 1
end;

function rnd (N : real) : real;      { Get a random number between N and N+1 }

begin
  rnd := random + N
end;

function tab (N : integer) : char;   { Tabulate to column N }

begin
  write(^M^['[',N);
  tab := 'C'
end;

procedure ClrEOS;		     { Clear to end of screen }

begin
  write(^['[J')
end;

function Key : char;		     { Wait for a key-press }

var
  C : char;

begin
  read(kbd,C);
  Key := C
end;

function UKey : char;		     { Get an upper-cased key }

begin
  UKey := upcase(Key)
end;

function Answer : char; 	     { Wait for player to press Y or N }

var
  C : char;

begin
  repeat
    C := UKey
  until C in ['Y','N'];
  Answer := C;
  if C = 'Y' then
    writeln('Yes.')
  else
    writeln('No.')
end;

function InputString : str_;   { Input a string. The following characters are
				 allowed: "A" to "Z", "a" to "z", "0" to "9"
				 and " ". Editing characters: <BACKSPACE> or
				 <DEL> - delete char, ^X - delete entire line,
				 <RETURN> - accept line. Maximum length = 40 }

var
  S : str_;
  C : char;
  I : integer;

begin
  S := '';
  repeat
    C := Key;
    case C of
      ^H, #127 : if S <> '' then
		   begin
		     delete(S,length(S),1);
		     write(^H' '^H)
		   end;
      ^X       : if S <> '' then
		   begin
		     for I := 1 to length(S) do
		       write(^H' '^H);
		     S := ''
		   end;
      ^M, ^J   : C := ^M
    else
      if (C in ['A'..'Z','a'..'z','0'..'9',' ']) and (length(S) < 40) then
	begin
	  S := S + C;
	  write(C)
	end
    end
  until C = ^M;
  InputString := S
end;

function InputNumber : real;	     { Input a positive real number. Same
				       editing characters as above. }

var
  Temp : real;
  S    : str_;
  C    : char;
  I    : integer;

begin
  S := '';
  repeat
    C := Key;
    case C of
      ^H, #127 : if S <> '' then
		   begin
		     delete(S,length(S),1);
		     write(^H' '^H)
		   end;
      ^X       : if S <> '' then
		   begin
		     for I := 1 to length(S) do
		       write(^H' '^H);
		     S := ''
		   end;
      ^M, ^J   : C := ^M
    else
      if (C in ['0'..'9','.']) and (length(S) < 18) then
	begin
	  S := S + C;
	  write(C)
	end
    end
  until C = ^M;
  if S = '' then
    S := '0';
  val(S,Temp,I);
  InputNumber := Temp
end;

procedure Title;	     { Display the title to screen }

begin
  ClrScr;
  GotoXY(28,1);
  writeln(^['[7;17m ***  Star Traders  *** '^['[17m');
  GotoXY(29,2);
  writeln('By J.N.Zaitseff, 1991.'^['[m');
  writeln
end;

procedure Continue;	     { Wait for the player to press any key }

var
  C : char;

begin
  write(^['[26C'^['[4;3;2;23m '^['[4;3;23m [ Press any key to continue ] ');
  write(^['[2m '^['[m ');
  C := Key
end;

{ Actual game procedures and functions start here. }

procedure InitData;	  { Initialise company data and map }

var
  X, Y : integer;

begin
  QuitSelected := false;
  randomize;
  Company[1].Name := 'Altair Starways';
  Company[2].Name := 'Betelgeuse, Ltd';
  Company[3].Name := 'Capella Freight Co';
  Company[4].Name := 'Denebola Shippers';
  Company[5].Name := 'Eridani Expediters';
  Company[6].Name := 'Fornax Express';
  Company[7].Name := 'Gemeni Inc';
  Company[8].Name := 'Hercules & Co';
  if not GameLoaded then
    begin
      for X := 1 to MaxCompanies do
	with Company[X] do
	  begin
	    SharePrice := 0.0;
	    Return := InitialReturn;
	    StockIssued := 0;
	    MaxStock := 0;
	    OnMap := false
	  end;
      InterestRate := InitialInterestRate;
      for X := 1 to MaxX do
	for Y := 1 to MaxY do
	  if rnd(0) < StarRatio then
	    Map[X,Y] := Star
	  else
	    Map[X,Y] := Empty
    end
end;

{$I-}

procedure Instructions;    { Display instructions to screen }

var
  C	   : char;
  F	   : text;

begin
  Title;
  assign(F,'TRADER.HLP');
  reset(F);
  if IOResult <> 0 then
    begin
      writeln;
      write(^['[21mHelp not available:'^['[m overlay file ');
      writeln(^['[17mTRADER.HLP'^['[m not on disk.');
      writeln;
      Continue
    end
  else
    begin
      while not eof(F) do
	begin
	  read(F,C);
	  case C of
	    ^F	: begin
		    Continue;
		    Title
		  end;
	    '|' : begin
		    read(F,C);
		    if C = '|' then
		      write(C)
		    else
		      write(^['[',C)
		  end;
	    '^' : begin
		    read(F,C);
		    case C of
		      '^'      : write(C);
		      'A'..'G' : write(^['[',ord(C)-48,'m');
		      'H'      : write(^['[19;7m')
		    end
		  end;
	    '~' : begin
		    read(F,C);
		    case C of
		      '~' : write(C);
		      'x' : write(MaxX:2);
		      'y' : write(MaxY:2);
		      'm' : write(NumberMoves:2);
		      'M' : write(chr(NumberMoves+$60));
		      'c' : write(MaxCompanies:2);
		      't' : write(MaxTurn:3)
		    end
		  end
	  else
	    write(C)
	  end
	end;
      close(F);
      Continue
    end
end;

{$I+}

procedure InitPlayers;	   { Initialise player data or load a game from disk }

var
  I, J : integer;
  C    : char;
  S    : str_;

{$I-}

procedure LoadGame (GN : integer);   { Load a game from disk }

var
  F : text;
  S : str_;
  K : integer;

begin
  S := 'TRADER'+chr(GN+$30)+'.GME';
  assign(F,S);
  reset(F);
  if IOResult <> 0 then
    begin
      writeln;
      writeln('        That game has not been saved to disk.');
      writeln;
      Continue;
      C := ' '
    end
  else
    begin
      readln(F,S);
      readln(F,S);
      if S <> Version then
	begin
	  close(F);
	  writeln;
	  write('	 That game was saved under an old version');
	  writeln(' of Star Traders.');
	  writeln;
	  Continue;
	  C := ' '
	end
      else
	begin
	  readln(F,MaxTurn);
	  readln(F,NumberPlayers);
	  for I := 1 to NumberPlayers do
	    with Player[I] do
	      begin
		readln(F,Name);
		readln(F,Cash);
		readln(F,Debt);
		readln(F,J);
		if J = 0 then
		  InGame := true
		else
		  InGame := false;
		for J := 1 to MaxCompanies do
		  readln(F,StockOwned[J])
	      end;
	  for I := 1 to MaxCompanies do
	    with Company[I] do
	      begin
		readln(F,SharePrice);
		readln(F,Return);
		readln(F,StockIssued);
		readln(F,MaxStock);
		readln(F,J);
		if J = 0 then
		  OnMap := true
		else
		  OnMap := false
	      end;
	  readln(F,InterestRate);
	  readln(F,TurnNumber);
	  readln(F,CurrentPlayer);
	  readln(F,FirstPlayer);
	  for I := 1 to MaxX do
	    for J := 1 to MaxY do
	      begin
		read(F,C);
		Map[I,J] := MapVal(ord(C)-$40)
	      end;
	  close(F);
	  C := 'C';
	  QuitSelected := false;
	  GameNumber := GN;
	  GameLoaded := true
	end
    end
end;

{$I+}

begin { InitPlayers }
  ClrScr;
  write(^['[>1h'^['[>2l'^['[>5l');
  GameLoaded := false;
  GameNumber := 0;
  if ParamCount > 0 then
    if ((ParamStr[1] <= '9') and (ParamStr[1] >= '1')) then
      begin
	C := ParamStr[1][1];
	Title;
	writeln;
	write('Loading game ',C,'... ');
	LoadGame(ord(C)-$30)
      end;
  if not GameLoaded then
    begin
      repeat
	Title;
	NumberPlayers := 0;
	write('Enter number of players ['^['[23m1'^['[m-');
	write(^['[23m',MaxPlayers,^['[m] ');
	write('or '^['[23m<C>'^['[m to continue a game: ');
	repeat
	  C := UKey;
	  NumberPlayers := ord(C)-$30
	until (NumberPlayers in [1..MaxPlayers]) or (C = 'C');
	writeln(C);
	if C = 'C' then
	  begin
	    writeln;
	    write('Enter game number ['^['[23m1'^['[m-');
	    write(^['[23m9'^['[m] or '^['[23m<ESC>'^['[m to cancel: ');
	    repeat
	      C := Key
	    until C in ['1'..'9',Esc];
	    if C <> Esc then
	      begin
		writeln(C);
		LoadGame(ord(C)-$30)
	      end
	  end
      until (C = 'C') or (NumberPlayers in [1..MaxPlayers]);
      if C <> 'C' then
	begin
	  MaxTurn := DefaultMaxTurn;
	  if ParamStr[1] = 'MAXTURN' then
	    begin
	      val(ParamStr[2],I,J);
	      if (J = 0) and (I >= 10) and (I <= 500) then
		MaxTurn := I
	    end;
	  for I := 1 to NumberPlayers do
	    begin
	      writeln;
	      write(^['[21mPlayer #',I,'.'^['[m Enter your name: ');
	      Player[I].Name := '';
	      repeat
		GotoXY(29,I+5);
		ClrEOL;
		Player[I].Name := InputString
	      until Player[I].Name <> '';
	      Player[I].Cash := InitialCash;
	      Player[I].Debt := 0.0;
	      Player[I].InGame := true;
	      for J := 1 to MaxCompanies do
		Player[I].StockOwned[J] := 0
	    end;
	  writeln;
	  writeln;
	  if NumberPlayers = 1 then
	    write('Do you need any ')
	  else
	    write('Does any player need ');
	  write('instructions? ['^['[23mY'^['[m/'^['[23mN'^['[m] ');
	  C := Answer;
	  if C = 'Y' then
	    Instructions;
	  Title;
	  writeln;
	  writeln(tab(24),'The game ends after '^['[17m',MaxTurn,^['[m moves.');
	  writeln;
	  Continue;
	  Title;
	  TurnNumber := 1;
	  CurrentPlayer := 1;
	  FirstPlayer := 1;
	  if NumberPlayers <> 1 then
	    begin
	      FirstPlayer := random(NumberPlayers)+1;
	      CurrentPlayer := FirstPlayer;
	      write('	     The first player to go is '^['[17m');
	      writeln(Player[FirstPlayer].Name,^['[m.');
	      writeln;
	      Continue
	    end
	end
    end
end;

procedure NextPlayer;	     { Get the next player }

var
  AllOut : boolean;
  I	 : integer;

begin
  AllOut := true;
  for I := 1 to NumberPlayers do
    if Player[I].InGame then
      AllOut := false;
  if not AllOut then
    repeat
      CurrentPlayer := CurrentPlayer + 1;
      if CurrentPlayer = NumberPlayers + 1 then
	CurrentPlayer := 1;
      if CurrentPlayer = FirstPlayer then
	TurnNumber := TurnNumber + 1
    until Player[CurrentPlayer].InGame
end;

procedure SelectMoves;	     { Select moves "a" to "MaxTurn" }

procedure SortMoves;	     { Sort moves from left to right }

var
  I, J, K : integer;
  Temp	  : MoveRec;

begin
  for I := 1 to NumberMoves - 1 do
    begin
      K := I;
      Temp := Move[I];
      for J := I+1 to NumberMoves do
	if Move[J].X < Temp.X then
	  begin
	    K := J;
	    Temp := Move[J]
	  end;
      Move[K] := Move[I];
      Move[I] := Temp
    end
end;

var
  I, J, TempX, TempY : integer;
  OK		     : boolean;

begin { SelectMoves }
  I := 0;
  for TempX := 1 to MaxX do
    for TempY := 1 to MaxY do
      if Map[TempX,TempY] = Empty then
	I := I + 1;
  if I < NumberMoves then
    QuitSelected := true
  else
    begin
      for I := 1 to NumberMoves do
	begin
	  repeat
	    TempX := random(MaxX)+1;
	    TempY := random(MaxY)+1;
	    OK := true;
	    if I > 1 then
	      for J := I-1 downto 1 do
		if ((Move[J].X = TempX) and (Move[J].Y = TempY)) then
		  OK := false
	  until OK and (Map[TempX,TempY] = Empty);
	  Move[I].X := TempX;
	  Move[I].Y := TempY
	end;
      QuitSelected := false;
      SortMoves
    end
end;

procedure DisplayMap;		 { Display the map to screen in colour }

const
  DefaultColour = ^['[22m';	 { Colour of empty space (blue) }

var
  X, Y, I : integer;

begin
  Title;
  write('Current Player: '^['[23m',Player[CurrentPlayer].Name,^['[m. ');
  write('Turn #'^['[23m',TurnNumber,^['[m');
  if TurnNumber = MaxTurn then
    write(^['[17;5m *** Last turn ***'^['[m');
  writeln;
  writeln(DefaultColour);
  for Y := 1 to MaxY do
    begin
      for X := 1 to MaxX do
	begin
	  if MaxX < 40 then
	    write(' ');
	  case Map[X,Y] of
	    Empty   : write('.');
	    Outpost : write(^['[m+',DefaultColour);
	    Star    : write(^['[21m*',DefaultColour)
	  else
	    write(^['[17m',chr(ord(Map[X,Y])-ord(C_A)+$41),DefaultColour)
	  end
	end;
      writeln
    end
end;

function TotalValue (N : integer) : real;	{ Calculate the total worth
						  of player N }

var
  I    : integer;
  Temp : real;

begin
  Temp := Player[N].Cash - Player[N].Debt;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      Temp := Temp + Player[N].StockOwned[I]*Company[I].SharePrice;
  TotalValue := Temp
end;

procedure DisplayStatus;	    { Display the current player's status }

var
  Temp	    : real;
  I	    : integer;
  AllOffMap : boolean;

begin
  Title;
  write('Player: '^['[23m',Player[CurrentPlayer].Name,^['[m.');
  writeln('	   Stock owned.');
  writeln;
  if TotalValue(CurrentPlayer) = 0 then
    writeln(^['[17;5m*** Bankrupt ***'^['[m')
  else
    begin
      AllOffMap := true;
      for I := 1 to MaxCompanies do
	if Company[I].OnMap then
	  AllOffMap := false;
      if AllOffMap then
	writeln('No companies on the map.')
      else
	begin
	  write(^['[23m');
	  writeln('			   Price per		   Holdings   Ownership of');
	  writeln('Company		   share ($)   Return (%)  (shares)    company (%)');
	  writeln(^['[m');
	  for I := 1 to MaxCompanies do
	    if Company[I].OnMap then
	      begin
		write(Company[I].Name,tab(24));
		write(Company[I].SharePrice:9:2,'    ');
		write(Company[I].Return*100:10:2,'  ');
		write(Player[CurrentPlayer].StockOwned[I]:8,'	');
		if Company[I].StockIssued = 0 then
		  Temp := 0.0
		else
		  Temp := Player[CurrentPlayer].StockOwned[I]
			  / Company[I].StockIssued;
		writeln(Temp*100:12:2)
	      end
	end;
      writeln;
      writeln('Current cash: '^['[23m$',Player[CurrentPlayer].Cash:4:2,^['[m');
      if Player[CurrentPlayer].Debt <> 0 then
	begin
	  writeln('Current debt: '^['[23m$',Player[CurrentPlayer].Debt:4:2,^['[m');
	  writeln('Debt interest rate: '^['[23m',InterestRate*100:4:2,'%'^['[m')
	end;
      writeln;
      writeln('Total value: '^['[21m$',TotalValue(CurrentPlayer):4:2,^['[m')
    end;
  writeln;
  Continue
end;

procedure GetMove;	      { Wait for player to enter their move }

var
  I, X, Y : integer;
  C	  : char;

begin
  if QuitSelected then
    Selection := Esc
  else
    begin
      repeat
	DisplayMap;
	write(^['[7;19m');
	for I := 1 to NumberMoves do
	  begin
	    X := Move[I].X;
	    Y := Move[I].Y;
	    GotoXY(X*2,Y+5);
	    write(chr(I+$60));
	  end;
	write(^['[m');
	repeat
	  GotoXY(1,22);
	  ClrEOS;
	  writeln;
	  write(^['[17m<1>'^['[m Display stock owned.  ');
	  writeln(^['[17m<2>'^['[m Declare bankruptcy.');
	  write(^['[17m<3>'^['[m Save game.	       ');
	  writeln(^['[17m<ESC>'^['[m Quit the game.');
	  GotoXY(1,22);
	  write('Select move ['^['[23ma'^['[m-'^['[23m');
	  write(chr(NumberMoves+$60),^['[m/'^['[23m1'^['[m-'^['[23m3');
	  write(^['[m/'^['[23m<ESC>'^['[m]: ');
	  repeat
	    C := UKey;
	    if C in ['1'..'3',Esc] then
	      I := 1
	    else
	      I := ord(C)-$40
	  until I in [1..NumberMoves];
	  if C = Esc then
	    writeln('<ESC>')
	  else
	    writeln(C);
	  ClrEOS;
	  Selection := C;
	  if Selection = '1' then
	    DisplayStatus
	  else
	    begin
	      write('Are you sure? ['^['[23mY'^['[m/'^['[23mN'^['[m] ');
	      C := Answer
	    end
	until C in ['Y','1']
      until C <> '1'
    end
end;

procedure ProcessMove;	      { Process the move the player selected }

var
  Left, Right, Up, Down, Temp : MapVal;
  X, Y, I		      : integer;
  AllOut		      : boolean;

procedure AssignSurroundVals;	      { What surrounds the X,Y position? }

begin
  if X = 1    then Left  := Empty else Left  := Map[X-1,Y];
  if X = MaxX then Right := Empty else Right := Map[X+1,Y];
  if Y = 1    then Up	 := Empty else Up    := Map[X,Y-1];
  if Y = MaxY then Down  := Empty else Down  := Map[X,Y+1]
end;

procedure IncreaseSharePrice (N : integer; P : real);	 { Increase the share
							   price and return }
var
  I : integer;

begin
  I := round(P);
  Company[N].SharePrice := Company[N].SharePrice + P + (P * rnd(0) / 2);
  Company[N].MaxStock := Company[N].MaxStock + (I div (random(10)+5));
  if rnd(0) < 0.25 then
    Company[N].Return := Company[N].Return * rnd(0.60)
end;

{$A-}

procedure IncludeOutpost (Xn, Yn : integer);	 { Include any outposts into
						   company. (Recursive) }
var
  L, R, U, D : MapVal;

begin
  Map[Xn,Yn] := Temp;
  IncreaseSharePrice(I,70.00);
  if Xn = 1    then L := Empty else L := Map[Xn-1,Yn];
  if Xn = MaxX then R := Empty else R := Map[Xn+1,Yn];
  if Yn = 1    then U := Empty else U := Map[Xn,Yn-1];
  if Yn = MaxY then D := Empty else D := Map[Xn,Yn+1];
  if L = Outpost then IncludeOutpost(Xn-1,Yn);
  if R = Outpost then IncludeOutpost(Xn+1,Yn);
  if U = Outpost then IncludeOutpost(Xn,Yn-1);
  if D = Outpost then IncludeOutpost(Xn,Yn+1)
end;

{$A+}

procedure EstablishNewCompany;	       { If a position next to an outpost or
					 star is selected }
var
  I, J	  : integer;
  AllUsed : boolean;
  Temp	  : MapVal;

const
  CompanyFormers : set of MapVal = [Outpost,Star];

begin
  if (Left in CompanyFormers) or (Right in CompanyFormers) or
     (Up   in CompanyFormers) or (Down	in CompanyFormers) then
    begin
      AllUsed := true;
      for I := MaxCompanies downto 1 do
	if not Company[I].OnMap then
	  begin
	    AllUsed := false;
	    J := I
	  end;
      if AllUsed then
	Map[X,Y] := Outpost
      else
	begin
	  Title;
	  writeln;
	  write(' A new company has been formed.');
	  writeln(' Its name is '^['[17m',Company[J].Name,^['[m.');
	  writeln;
	  writeln;
	  Continue;
	  Temp := MapVal(J+ord(C_A)-1);
	  Map[X,Y] := Temp;
	  Company[J].Return := InitialReturn;
	  Company[J].SharePrice := 60.00;
	  Company[J].OnMap := true;
	  Company[J].StockIssued := 5;
	  Company[J].MaxStock := 50;
	  for I := 1 to NumberPlayers do
	    Player[I].StockOwned[J] := 0;
	  Player[CurrentPlayer].StockOwned[J] := 5
	end
    end
end;

procedure MergeCompanies (Company1, Company2 : MapVal);   { Perform a merge }

var
  I, J, X1, Y1	    : integer;
  Bonus, Val1, Val2 : real;
  ExtraStock	    : integer;
  SwapTemp	    : MapVal;

begin
  I := ord(Company1)-ord(C_A)+1;
  J := ord(Company2)-ord(C_A)+1;
  Val1 := Company[I].SharePrice * Company[I].StockIssued * Company[I].Return;
  Val2 := Company[J].SharePrice * Company[J].StockIssued * Company[J].Return;
  if Val1 < Val2 then
    begin
      SwapTemp := Company1;
      Company1 := Company2;
      Company2 := SwapTemp;
      X1 := I;
      I := J;
      J := X1
    end;
  Title;
  write(^['[17m',Company[J].Name,^['[m has just merged into '^['[17m');
  writeln(Company[I].Name,^['[m.');
  writeln;
  writeln('Please note the following transactions:');
  writeln;
  writeln('Old stock: '^['[23m',Company[J].Name,^['[m.');
  writeln('New stock: '^['[23m',Company[I].Name,^['[m.');
  writeln;
  write(^['[23mPlayer',tab(25),'Old stock   New stock	Total holdings');
  writeln('  Bonus paid ($)'^['[m');
  writeln;
  Y1 := 0;
  for X1 := 1 to NumberPlayers do
    if Player[X1].InGame then
      begin
	write(Player[X1].Name,tab(25));
	ExtraStock := Player[X1].StockOwned[J] div 2;
	write(Player[X1].StockOwned[J]:9,'   ',ExtraStock:9,'	');
	Player[X1].StockOwned[I] := Player[X1].StockOwned[I] + ExtraStock;
	Y1 := Y1 + ExtraStock;
	write(Player[X1].StockOwned[I]:14,'  ');
	if Company[J].StockIssued = 0 then
	  Bonus := 0.00
	else
	  Bonus := 10 * ((Player[X1].StockOwned[J] / Company[J].StockIssued)
		   * Company[J].SharePrice);
	writeln(Bonus:14:2);
	Player[X1].Cash := Player[X1].Cash + Bonus;
	Player[X1].StockOwned[J] := 0
      end;
  writeln;
  Continue;
  Map[X,Y] := Company2;
  with Company[I] do
    begin
      StockIssued := StockIssued + Y1;
      SharePrice := SharePrice + Company[J].SharePrice / rnd(1.5);
      MaxStock := MaxStock + Y1
    end;
  with Company[J] do
    begin
      StockIssued := 0;
      MaxStock := 0;
      OnMap := false
    end;
  for X1 := 1 to MaxX do
    for Y1 := 1 to MaxY do
      if Map[X1,Y1] = Company2 then
	Map[X1,Y1] := Company1;
  AssignSurroundVals
end;

procedure SaveGame;	      { Save the game to disk }

var
  F : text;
  C : char;
  S : str_;

begin
  Title;
  if GameLoaded then
    C := chr(GameNumber+$30)
  else
    begin
      writeln;
      writeln;
      write('Enter game number to save ['^['[23m1'^['[m-'^['[23m9'^['[m]: ');
      repeat
	C := Key
      until C in ['1'..'9'];
      writeln(C)
    end;
  writeln;
  writeln;
  write('Saving game... ');
  S := 'TRADER'+C+'.GME';
  assign(F,S);
  rewrite(F);
  writeln(F,'@@@ Star Traders');
  writeln(F,Version);
  writeln(F,MaxTurn);
  writeln(F,NumberPlayers);
  for I := 1 to NumberPlayers do
    with Player[I] do
      begin
	writeln(F,Name);
	writeln(F,Cash);
	writeln(F,Debt);
	if InGame then
	  writeln(F,'0')
	else
	  writeln(F,'1');
	for X := 1 to MaxCompanies do
	  writeln(F,StockOwned[X])
      end;
  for I := 1 to MaxCompanies do
    with Company[I] do
      begin
	writeln(F,SharePrice);
	writeln(F,Return);
	writeln(F,StockIssued);
	writeln(F,MaxStock);
	if OnMap then
	  writeln(F,'0')
	else
	  writeln(F,'1')
      end;
  writeln(F,InterestRate);
  writeln(F,TurnNumber);
  writeln(F,CurrentPlayer);
  writeln(F,FirstPlayer);
  for X := 1 to MaxX do
    for Y := 1 to MaxY do
      write(F,chr(ord(Map[X,Y])+$40));
  writeln(F);
  writeln(F,'@@@ Star Traders');
  close(F);
  Selection := Esc
end;

begin { ProcessMove }
  Title;
  if Selection = '2' then	   { Declare bankruptcy }
    begin
      write(^['[17m',Player[CurrentPlayer].Name);
      writeln(^['[m has declared bankruptcy.');
      writeln;
      Continue;
      DisplayStatus;
      Title;
      Player[CurrentPlayer].InGame := false;
      for I := 1 to MaxCompanies do
	begin
	  Company[I].MaxStock := Company[I].MaxStock +
				 Player[CurrentPlayer].StockOwned[I];
	  Company[I].StockIssued := Company[I].StockIssued -
				    Player[CurrentPlayer].StockOwned[I];
	  Player[CurrentPlayer].StockOwned[I] := 0
	end;
      Player[CurrentPlayer].Cash := 0.00;
      Player[CurrentPlayer].Debt := 0.00;
      AllOut := true;
      for I := 1 to NumberPlayers do
	if Player[I].InGame then
	  AllOut := false;
      if AllOut then
	Selection := Esc
    end;
  if Selection = '3' then	  { Save the game }
    SaveGame;
  if Selection = Esc then	  { Quit the game }
    QuitSelected := true
  else
    if Player[CurrentPlayer].InGame then
      begin
	I := ord(Selection) - $40;
	X := Move[I].X;
	Y := Move[I].Y;
	AssignSurroundVals;
	if (Left = Empty) and (Right = Empty) and (Up = Empty) and (Down = Empty) then
	  Map[X,Y] := Outpost;
	if (Left in CompanyVal) and (Right in CompanyVal) and (Left <> Right) then
	  MergeCompanies(Left,Right);
	if (Left in CompanyVal) and (Up in CompanyVal) and (Left <> Up) then
	  MergeCompanies(Left,Up);
	if (Left in CompanyVal) and (Down in CompanyVal) and (Left <> Down) then
	  MergeCompanies(Left,Down);
	if (Right in CompanyVal) and (Up in CompanyVal) and (Right <> Up) then
	  MergeCompanies(Right,Up);
	if (Right in CompanyVal) and (Down in CompanyVal) and (Right <> Down) then
	  MergeCompanies(Right,Down);
	if (Up in CompanyVal) and (Down in CompanyVal) and (Up <> Down) then
	  MergeCompanies(Up,Down);
	if not((Left in CompanyVal) or (Right in CompanyVal) or (Up in CompanyVal)
	   or (Down in CompanyVal)) then
	  EstablishNewCompany;
	Temp := Empty;
	if Left  in CompanyVal then Temp := Left;
	if Right in CompanyVal then Temp := Right;
	if Up	 in CompanyVal then Temp := Up;
	if Down  in CompanyVal then Temp := Down;
	if Temp <> Empty then
	  begin
	    I := ord(Temp)-ord(C_A)+1;
	    Map[X,Y] := Temp;
	    IncreaseSharePrice(I,60.00)
	  end;
	Temp := Map[X,Y];
	I := 0;
	if Temp in CompanyVal then I := ord(Temp)-ord(C_A)+1;
	if I > 0 then
	  begin
	    if Left  = Star then IncreaseSharePrice(I,300.00);
	    if Right = Star then IncreaseSharePrice(I,300.00);
	    if Up    = Star then IncreaseSharePrice(I,300.00);
	    if Down  = Star then IncreaseSharePrice(I,300.00);
	  end;
	if Temp in CompanyVal then
	  begin
	    if Left  = Outpost then IncludeOutpost(X-1,Y);
	    if Right = Outpost then IncludeOutpost(X+1,Y);
	    if Up    = Outpost then IncludeOutpost(X,Y-1);
	    if Down  = Outpost then IncludeOutpost(X,Y+1)
	  end
      end
end;

procedure StockExchange;      { Enter the Interstellar Stock Exchange }

var
  AllOff  : boolean;
  I, J	  : integer;
  C	  : char;
  BidUsed : boolean;
  R, R2   : real;

procedure AdjustValues;       { Adjust the cost of shares, their return, the
				interest rate, etc. }

var
  X, Y : integer;
  Temp : MapVal;

begin
  if rnd(0) > 0.99 then
    begin
      I := random(MaxCompanies)+1;
      if Company[I].OnMap then
	begin
	  Title;
	  writeln;
	  write(^['[17m',Company[I].Name,^['[m has been declared bankrupt ');
	  writeln('by the Interstellar Trading Bank.');
	  writeln;
	  R := 0;
	  if rnd(0) > 0.2 then
	    R := rnd(0);
	  if R = 0 then
	    writeln('All assets have been taken to repay outstanding loans.')
	  else
	    begin
	      write('The bank has agreed to pay stock holders '^['[23m');
	      writeln(R*100:4:2,^['[m% of the share value on each share owned.');
	      writeln;
	      write('Old share value:	    '^['[23m$');
	      writeln(Company[I].SharePrice:4:2,^['[m');
	      write('Amount paid per share: '^['[23m$');
	      writeln(Company[I].SharePrice * R:4:2,^['[m');
	      writeln(^['[23m');
	      writeln('Player',tab(25),'Holdings   Value ($)   Amount paid ($)');
	      writeln(^['[m');
	      for J := 1 to NumberPlayers do
		if Player[J].InGame then
		  begin
		    write(Player[J].Name,tab(25),Player[J].StockOwned[I]:8);
		    R2 := Player[J].StockOwned[I] * Company[I].SharePrice;
		    writeln('	',R2:9:2,'   ',R2 * R:15:2);
		    Player[J].Cash := Player[J].Cash + R2 * R
		  end
	    end;
	  writeln;
	  Continue;
	  for J := 1 to NumberPlayers do
	    Player[J].StockOwned[I] := 0;
	  with Company[I] do
	    begin
	      SharePrice := 0.00;
	      Return := 0.00;
	      StockIssued := 0;
	      MaxStock := 0;
	      OnMap := false
	    end;
	  Temp := MapVal(I+ord(C_A)-1);
	  for X := 1 to MaxX do
	    for Y := 1 to MaxY do
	      if Map[X,Y] = Temp then
		Map[X,Y] := Empty
	end
    end;
  if rnd(0) < 0.4 then
    begin
      I := random(MaxCompanies)+1;
      Company[I].Return := Company[I].Return * rnd(0.75)
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap and (Company[I].Return > 0.40) then
      Company[I].Return := Company[I].Return / rnd(1.5);
  if rnd(0) < 0.3 then
    begin
      I := random(MaxCompanies)+1;
      R := rnd(0) * (Company[I].SharePrice / 4);
      if rnd(0) < 0.65 then
	R := -R;
      Company[I].SharePrice := Company[I].SharePrice + R
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      begin
	if Company[I].StockIssued = 0 then
	  R := 0
	else
	  R := Player[CurrentPlayer].StockOwned[I] / Company[I].StockIssued;
	with Player[CurrentPlayer] do
	  Cash := Cash + (StockOwned[I] * Company[I].SharePrice *
		  Company[I].Return) + (R * Company[I].SharePrice * 2)
      end;
  if rnd(0) < 0.3 then
    InterestRate := InterestRate * rnd(0.65);
  if InterestRate > 0.30 then
    InterestRate := InterestRate / rnd(1.5);
  if Player[CurrentPlayer].Debt * InterestRate > 10000.00 then
    R := 0.5
  else
    R := 1;
  with Player[CurrentPlayer] do
    Debt := Debt * (1 + InterestRate * R)
end;

procedure Title2;

begin
  Title;
  GotoXY(22,4);
  writeln(^['[4;21m***	Interstellar Stock Exchange  ***'^['[m');
  writeln('Player: '^['[23m',Player[CurrentPlayer].Name,^['[m')
end;

procedure VisitBank;		{ Visit the Interstellar Trading Bank }

var
  R, R2 : real;
  C	: char;

begin
  Title2;
  writeln(tab(22),^['[7m === Interstellar Trading Bank === '^['[m');
  writeln;
  writeln('Current cash: '^['[23m$',Player[CurrentPlayer].Cash:4:2,^['[m');
  writeln('Current debt: '^['[23m$',Player[CurrentPlayer].Debt:4:2,^['[m');
  writeln('Interest rate: '^['[23m',InterestRate * 100:4:2,'%'^['[m');
  R2 := 2 * (TotalValue(CurrentPlayer) - Player[CurrentPlayer].Debt);
  if R2 < 0.00 then
    R2 := 0.00;
  writeln('Credit limit: '^['[23m$',R2:4:2,^['[m');
  writeln;
  writeln(^['[17m<1>'^['[m Borrow money.');
  writeln(^['[17m<2>'^['[m Repay debt.');
  writeln(^['[17m<3>'^['[m Exit from Bank.');
  writeln;
  write('Enter selection ['^['[23m1'^['[m/'^['[23m2'^['[m/'^['[23m3'^['[m]: ');
  repeat
    C := Key
  until C in ['1'..'3',' '];
  writeln(C);
  writeln;
  case C of
    '1' : if R2 = 0 then
	    begin
	      writeln('The bank will not lend you any more money.');
	      writeln;
	      Continue
	    end
	  else
	    begin
	      write('How much do you wish to borrow? ');
	      repeat
		GotoXY(33,19);
		ClrEOL;
		R := InputNumber
	      until R <= (R2 + 0.01);
	      with Player[CurrentPlayer] do
		begin
		  Cash := Cash + R;
		  Debt := (1+ InterestRate) * (Debt + R)
		end
	    end;
    '2' : if Player[CurrentPlayer].Debt = 0.00 then
	    begin
	      writeln('You have no debt to pay off.');
	      writeln;
	      Continue
	    end
	  else
	    begin
	      write('How much do you wish to repay? ');
	      repeat
		GotoXY(32,19);
		ClrEOL;
		R := InputNumber
	      until (R <= (Player[CurrentPlayer].Cash + 0.01)) and
		    (R <= (Player[CurrentPlayer].Debt + 0.01));
	      with Player[CurrentPlayer] do
		begin
		  Cash := Cash - R;
		  Debt := Debt - R;
		  if Cash < 0.01 then
		    Cash := 0.00;
		  if Debt < 0.01 then
		    Debt := 0.00
		end
	    end
  end
end;

procedure Trade (N : integer);	  { Buy, sell or bid for shares with company N }

var
  R, R2 : real;
  C	: char;
  I, J	: integer;

procedure Title3;

begin
  Title2;
  writeln;
  writeln(^['[22m--- Stock transaction in ',Company[N].Name,' ---'^['[m');
  writeln;
  writeln('Current cash:     '^['[23m$',Player[CurrentPlayer].Cash:4:2,^['[m');
  write('Current holdings: '^['[23m');
  writeln(Player[CurrentPlayer].StockOwned[N],^['[m');
  writeln;
  writeln('Each share costs '^['[23m$',Company[N].SharePrice:4:2,^['[m');
  write('There are '^['[23m',Company[N].MaxStock - Company[N].StockIssued);
  writeln(^['[m shares left.')
end;

begin { Trade }
  Title2;
  writeln;
  with Company[N] do
    begin
      writeln('Company name:  '^['[23m',Name,^['[m');
      writeln('Shares issued: '^['[23m',StockIssued,^['[m');
      writeln('Shares left:   '^['[23m',MaxStock - StockIssued,^['[m');
      writeln('Price:	      '^['[23m$',SharePrice:4:2,^['[m');
      writeln('Return:	      '^['[23m',Return*100:4:2,'%'^['[m')
    end;
  writeln;
  writeln('Current cash:     '^['[23m$',Player[CurrentPlayer].Cash:4:2,^['[m');
  write('Current holdings: '^['[23m');
  writeln(Player[CurrentPlayer].StockOwned[N],^['[m');
  if Company[N].StockIssued = 0 then
    R2 := 0.0
  else
    R2 := Player[CurrentPlayer].StockOwned[N] / Company[N].StockIssued;
  writeln('Percentage owned: '^['[23m',R2*100:4:2,'%'^['[m');
  writeln;
  write(^['[17m<1>'^['[m Buy stock.',tab(45));
  writeln(^['[17m<2>'^['[m Sell stock.');
  write(^['[17m<3>'^['[m Bid company to issue more shares.',tab(45));
  writeln(^['[17m<4>'^['[m Exit.');
  writeln;
  write('Enter selection ['^['[23m1'^['[m-'^['[23m4'^['[m]: ');
  repeat
    C := Key
  until C in ['1'..'4',' '];
  writeln(C);
  case C of
    '1' : begin
	    Title3;
	    R := Player[CurrentPlayer].Cash / Company[N].SharePrice;
	    if R > 32000 then
	      R := 32000;
	    I := trunc(R);
	    with Company[N] do
	      if I > (MaxStock - StockIssued) then
		I := MaxStock - StockIssued;
	    writeln('You can purchase '^['[23m',I,^['[m shares.');
	    writeln;
	    write('How many shares do you wish to purchase? ');
	    repeat
	      GotoXY(42,16);
	      ClrEOL;
	      R := InputNumber;
	      if R > 32000 then
		R := 32000;
	      J := trunc(R)
	    until (J >= 0) and (J <= I);
	    with Player[CurrentPlayer] do
	      begin
		StockOwned[N] := StockOwned[N] + J;
		Company[N].StockIssued := Company[N].StockIssued + J;
		Cash := Cash - (J * Company[N].SharePrice)
	      end
	  end;
    '2' : if Player[CurrentPlayer].StockOwned[N] = 0 then
	    begin
	      Title2;
	      writeln;
	      writeln('You do not have any shares to sell.');
	      writeln;
	      Continue
	    end
	  else
	    begin
	      Title3;
	      writeln;
	      write('How many shares do you wish to sell? ');
	      repeat
		GotoXY(38,15);
		ClrEOL;
		R := InputNumber;
		if R > 32000 then
		  R := 32000;
		J := trunc(R)
	      until (J >= 0) and (J <= Player[CurrentPlayer].StockOwned[N]);
	      with Player[CurrentPlayer] do
		begin
		  StockOwned[N] := StockOwned[N] - J;
		  Company[N].StockIssued := Company[N].StockIssued - J;
		  Cash := Cash + (J * Company[N].SharePrice)
		end
	    end;
    '3' : begin
	    R := 0;
	    if (not BidUsed) and (rnd(0) < R2) and (rnd(0) < 0.75) then
	      begin
		R := rnd(0) * R2 * 200;
		Company[N].MaxStock := Company[N].MaxStock + trunc(R)
	      end;
	    BidUsed := true;
	    Title3;
	    writeln;
	    write(^['[23m',Company[N].Name,^['[m has ');
	    if trunc(R) = 0 then
	      writeln('refused to issue more shares.')
	    else
	      writeln('issued '^['[23m',trunc(R),^['[m more shares.');
	    writeln;
	    Continue
	  end
  end
end;

var
  OK : boolean;

begin { StockExchange }
  if (not QuitSelected) and Player[CurrentPlayer].InGame then
    begin
      BidUsed := false;
      AdjustValues;
      if TotalValue(CurrentPlayer) <= -100000.00 then
	begin
	  Title;
	  writeln;
	  write('Your debt has amounted to '^['[17m$');
	  writeln(Player[CurrentPlayer].Debt:4:2,^['[m');
	  writeln;
	  write('The bank has impounded '^['[23m$');
	  writeln(Player[CurrentPlayer].Cash:4:2,^['[m from your cash.');
	  writeln;
	  Continue;
	  with Player[CurrentPlayer] do
	    begin
	      Debt := Debt / (InterestRate + 1) - Cash;
	      Cash := 0.00;
	      if Debt < 0.00 then
		Debt := 0.00
	    end;
	  if (TotalValue(CurrentPlayer) <= 0.00) and (rnd(0) < 0.07) then
	    begin
	      Title;
	      writeln;
	      write(^['[21mYou have been declared bankrupt by the ');
	      writeln('Interstellar Trading Bank.'^['[m');
	      writeln;
	      Continue;
	      DisplayStatus;
	      Title;
	      Player[CurrentPlayer].InGame := false;
	      for I := 1 to MaxCompanies do
		with Company[I] do
		  begin
		    MaxStock := MaxStock + Player[CurrentPlayer].StockOwned[I];
		    StockIssued := StockIssued - Player[CurrentPlayer].StockOwned[I];
		    Player[CurrentPlayer].StockOwned[I] := 0
		  end;
	      Player[CurrentPlayer].Cash := 0.00;
	      Player[CurrentPlayer].Debt := 0.00;
	      AllOff := true;
	      for I := 1 to NumberPlayers do
		if Player[I].InGame then
		  AllOff := false;
	      if AllOff then
		QuitSelected := true
	    end
	end
    end;
  if (not QuitSelected) and Player[CurrentPlayer].InGame then
    repeat
      Title2;
      writeln;
      AllOff := true;
      for I := 1 to MaxCompanies do
	if Company[I].OnMap then
	  AllOff := false;
      if AllOff then
	writeln('No companies on the map.')
      else
	begin
	  write(^['[23m  Company',tab(25),'Shares issued   Shares left	 ');
	  writeln('Price ($)   Return (%)'^['[m');
	  writeln;
	  for I := 1 to MaxCompanies do
	    if Company[I].OnMap then
	      with Company[I] do
		begin
		  write(^['[7;17m',Name[1],^['[m ');
		  write(Name,tab(25),StockIssued:13,'	');
		  write(MaxStock-StockIssued:11,'   ');
		  writeln(SharePrice:9:2,'   ',Return*100:10:2)
		end
	end;
      writeln;
      writeln;
      write(^['[17m<1>'^['[m Display stock owned.',tab(45));
      writeln(^['[17m<2>'^['[m Display map.');
      write(^['[17m<3>'^['[m Visit the Interstellar Trading Bank.',tab(45));
      writeln(^['[17m<4>'^['[m Exit the Stock Exchange.');
      writeln;
      write('Enter selection ['^['[17mCompany letter');
      write(^['[m/'^['[23m1'^['[m-'^['[23m4'^['[m]: ');
      repeat
	C := UKey;
	if C in ['1'..'4',' '] then
	  OK := true
	else
	  if (C >= 'A') and (C <= chr(MaxCompanies+$40)) then
	    if Company[ord(C)-$40].OnMap then
	      OK := true
	    else
	      OK := false
	  else
	    OK := false
      until OK;
      write(C);
      case C of
	'1'	 : DisplayStatus;
	'2'	 : begin
		     DisplayMap;
		     writeln;
		     Continue
		   end;
	'3'	 : VisitBank;
	'4', ' ' : C := ' '
      else
	Trade(ord(C)-$40)
      end
    until C = ' '
end;

procedure DisplayStatus_of_All;    { Display everyone's status at end of game }

var
  I, J : integer;
  C    : char;

procedure SortPlayers;		   { Sort players according to total value }

var
  I, J, K : integer;
  Temp	  : PlayerRec;
  R1, R2  : real;

begin
  for I := 1 to NumberPlayers - 1 do
    begin
      K := I;
      Temp := Player[I];
      for J := I+1 to NumberPlayers do
	if TotalValue(K) < TotalValue(J) then
	  begin
	    K := J;
	    Temp := Player[J]
	  end;
      Player[K] := Player[I];
      Player[I] := Temp
    end
end;

begin { DisplayStatus_of_all }
  Title;
  writeln;
  writeln;
  writeln('--- The game is over after '^['[23m',TurnNumber-1,^['[m turns ---');
  writeln;
  writeln;
  Continue;
  for I := 1 to NumberPlayers do
    begin
      CurrentPlayer := I;
      DisplayStatus
    end;
  Title;
  if NumberPlayers = 1 then
    writeln('Your total value was '^['[23m$',TotalValue(1):4:2,^['[m')
  else
    begin
      SortPlayers;
      write('The winner is '^['[17m',Player[1].Name,^['[m');
      if TotalValue(1) = 0.00 then
	writeln(', who is '^['[7;17m *** Bankrupt *** '^['[m.')
      else
	writeln(', with a value of '^['[23m$',TotalValue(1):4:2,^['[m');
      writeln;
      writeln(^['[23m');
      writeln('Position',tab(17),'Player',tab(40),'Total value ($)');
      writeln(^['[m');
      for I := 1 to NumberPlayers do
	begin
	  write(I);
	  case I of
	    1 : write('st');
	    2 : write('nd');
	    3 : write('rd')
	  else
	    write('th')
	  end;
	  write(tab(17),Player[I].Name,tab(40));
	  if TotalValue(I) = 0.00 then
	    writeln(^['[17;5m*** BANKRUPT ***'^['[m')
	  else
	    writeln(TotalValue(I):16:2)
	end
    end;
  writeln;
  writeln;
  write(^['[26C'^['[4;3;2;23m '^['[4;3;23m [ Press <RETURN> to exit ] ');
  write(^['[2m '^['[m ');
  repeat
    C := Key
  until C = ^M;
  ClrScr;
  write(^['[>1l'^['[>2h')
end;


begin  { Main }
  InitPlayers;
  InitData;
  repeat
    SelectMoves;
    GetMove;
    ProcessMove;
    StockExchange;
    NextPlayer
  until (TurnNumber = MaxTurn + 1) or QuitSelected;
  DisplayStatus_of_All
end.
