{**************************************************************************}
{*                                                                        *}
{*                          Computer Delay Routines                       *}
{*                                                                        *}
{*                          (C) J.N.Zaitseff, 1994.                       *}
{*                                                                        *}
{**************************************************************************}

{ This unit provides a general-purpose delay routine. }
{ December 1998:  This unit has been patched to allow correct operation on
  machines with fast processors (>= Pentium Pro 200MHz).  This corrects the
  "Runtime Error 200" issue. }

unit Delays;

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E-,F-,N-,O-,P-,T+,V+,X+}
{$IFDEF DPMI} {$G+,F-,O-} {$ELSE} {$G-,F+,O+} {$ENDIF}
{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

interface

procedure Delay (ms : word);
{ Generate a delay for the specified number of milliseconds. }

implementation

{$IFDEF USEFLOPWAIT}  { Only some very old machines required this unit }
uses
  FlopWait;
{$ENDIF}

var
  DelayCnt : LongInt;

type
  LongRec = record
              Lo, Hi : word
            end;


procedure DelayLoop; near; assembler;

asm
@@1:    sub     ax,1
        sbb     dx,0
        jc      @@2
        cmp     bl,es:[di]
        je      @@1
@@2:
end;

procedure Initialise; assembler;

asm
        mov     es,Seg0040
        mov     di,6Ch          { Offset of BIOS timer variable }
        mov     bl,es:[di]
@@1:    cmp     bl,es:[di]
        je      @@1
        mov     bl,es:[di]
        mov     ax,-28
        cwd
        call    DelayLoop
        not     ax
        not     dx
        mov     bx,ax
        mov     ax,dx
        xor     dx,dx
        mov     cx,55
        div     cx
        xchg    ax,bx
        div     cx
        mov     DelayCnt.LongRec.Lo,ax
        mov     DelayCnt.LongRec.Hi,bx
end;

procedure Delay (ms : word); assembler;

asm
	mov	cx,ms
	mov	es,Seg0040
	xor	di,di
	mov	bl,es:[di]
@@1:	mov	ax,DelayCnt.LongRec.Lo
	mov	dx,DelayCnt.LongRec.Hi
	call	DelayLoop
	loop	@@1
@@2:
end;

begin
  Initialise
end.
