{**************************************************************************}
{*                                                                        *}
{*                        Speaker Sounds Routines                         *}
{*                                                                        *}
{*                       (C) J.N.Zaitseff, 1994-98.                       *}
{*                                                                        *}
{**************************************************************************}

{ This unit provides a program with the ability to generate sounds from the
  internal PC speaker. }
{ December 1998:  This unit has been modified to directly use the Delays
  unit instead of containing the Delay routine itself }


unit Speaker;

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E-,F-,N-,O-,P-,T+,V+,X+}
{$IFDEF DPMI} {$G+,F-,O-} {$ELSE} {$G-,F+,O+} {$ENDIF}
{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

interface

procedure Beep;
{ Generate a generic "beep" from the speaker. This is suitable for warnings,
  etc. in programs. }

procedure Sound (Hz : word);
{ Start a sound of the specified frequency (in Hertz). Call NoSound to stop
  the sound. Use Delay for a specific delay. Hence a typical sequence might
  be:
      Sound(1000);
      Delay(1000);
      NoSound;
  This generates a 1000Hz note for one second. }

procedure NoSound;
{ Turn sound off. }


implementation

uses
{$IFDEF USEFLOPWAIT}
  FlopWait,
{$ENDIF}
  Delays;


procedure Sound (Hz : word); assembler;

asm
	mov	bx,Hz
	mov	ax,34DDh
	mov	dx,0012h
	cmp	dx,bx
	jnc	@@2
	div	bx
	mov	bx,ax
	in	al,61h
	test	al,3
	jnz	@@1
	or	al,3
	out	61h,al
	mov	al,0B6h
	out	43h,al
@@1:	mov	al,bl
	out	42h,al
	mov	al,bh
	out	42h,al
@@2:
end;

procedure NoSound; assembler;

asm
	in	al,61h
	and	al,0FCh
	out	61h,al
end;

procedure Beep;

begin
  Sound(880);
  Delay(100);
  NoSound
end;

end.
