{************************************************************************}
{*                                                                      *}
{*               Text-oriented keyboard input line routines.            *}
{*                                                                      *}
{*                          (C) J.N.Zaitseff, 1993.                     *}
{*                                                                      *}
{************************************************************************}

unit TxInput;

{$A+,B-,D-,F+,G-,N-,O+,Q-,R-,S-}

interface

uses
  Keyb;

const
                        { Useful constants for DataSet }
  ASCIIdata     = [' '..'~'];
  Alphanumerics = [' ','A'..'Z','a'..'z','0'..'9'];
  NumericData   = ['0'..'9','E','.','-'];
  FilenameData  = ['A'..'Z','0'..'9','.',':','\'];

type
  InputResultCode =     { Result codes for InputLine }
      (inp_Normal,                { <ENTER> pressed }
       inp_ExtKeyPressed,         { Extended or control key pressed }
       inp_LeftButtonPressed,     { Left mouse button pressed }
       inp_RightButtonPressed);   { Right mouse button pressed }

  InputRec = record
               DataSet     : set of char;  { * Allowable characers }
               MakeUpper   : boolean;      { * Make char into upper case }
               Default     : string;       { * Default editing string }
               UseDefault  : boolean;      { * Use the default string or not }
               NewEdit     : boolean;      { * TRUE for new edit, FALSE if edit
                                             continues from before }
               MaxLen      : integer;      { * Maximum string length }
               FCol, BCol  : integer;      { * Foreground, background colour }
               X, Y        : integer;      { * Starting X, Y coordinate }
               XPos        : integer;      { Used internally }
               SPos        : integer;      { Used internally }
               ExtKey      : KeyStr;       { Extended / control key return }
               MouseX,                     { Mouse coordinate of click }
               MouseY      : integer;
             end;                          { NB: (*) means set prior to call }

const
  DefaultInputRec : InputRec = (DataSet    : ASCIIdata;
                                MakeUpper  : false;
                                Default    : '';
                                UseDefault : false;
                                NewEdit    : true);

const
  InsMode : boolean = true;     { TRUE = insert mode, FALSE = overtype mode }


procedure InputString (var St : string; var InputParam : InputRec;
                       var Result : InputResultCode);
{ Input a string in text mode. St = result of edit, InputParam is the
  associated record (some fields must be set prior to call), Result is
  returned on exit. }

procedure InputPassword (var St : string; var InputParam : InputRec;
                         var Result : InputResultCode);
{ As above, except that the string is not echoed. }


implementation

uses
  Mouse, Crt, Speaker;

procedure InputString (var St : string; var InputParam : InputRec;
                       var Result : InputResultCode);

var
  OldX, OldY,
  OldC,
  OldWMin,
  OldWMax      : word;
  Editing      : boolean;
  K            : KeyStr;
  CursPos      : integer;
  MX, MY       : integer;
  MBPressed    : boolean;

procedure SetCursor; assembler;

asm
        int     11h             { Get equipment list }
        mov     cl,4
        shr     ax,cl
        and     al,3
        mov     cx,7
        cmp     al,3            { CGA mode? }
        jnz     @1              { No:  7 scan lines for cursor }
        mov     cl,13           { Yes: 13 scan lines for cursor }
@1:     cmp     InsMode,true
        jnz     @2
        mov     ch,cl
        dec     ch
@2:     mov     ah,1            { Set cursor }
        int     10h
end;

procedure DisplayText;

begin
  with InputParam do
    begin
      HidePointer;
      TextColor(FCol);
      TextBackground(BCol);
      ClrScr;
      GotoXY(1,1);
      write(St);
      RestorePointer
    end
end;

procedure InitEdit;

begin
  with InputParam do
    begin
      Editing := true;
      LeftButtonPressCount(MX,MY);
      RightButtonPressCount(MX,MY);
      if NewEdit then
        SPos := 1;
      CursPos := SPos;
      Window(X+lo(OldWMin),Y+hi(OldWMin),X+lo(OldWMin)+MaxLen,Y+hi(OldWMin));
      if UseDefault then
        St := Default;
      if length(St) > MaxLen then
        St[0] := chr(MaxLen);
      DisplayText
    end
end;

procedure Abort_common;

begin
  with InputParam do
    begin
      Editing := false;
      NewEdit := false;
      UseDefault := true;
      Default := St;
      SPos := CursPos
    end
end;

procedure AbortEditing;

begin
  Result := inp_ExtKeyPressed;
  InputParam.ExtKey := K;
  Abort_common
end;

procedure MouseAbort;

begin
  with InputParam do
    begin
      MouseX := MX;
      MouseY := MY
    end;
  Abort_common
end;

procedure DeleteLeft;

begin
  if CursPos = 1 then
    Beep
  else
    begin
      dec(CursPos);
      delete(St,CursPos,1);
      DisplayText
    end
end;

procedure DeleteChar;

begin
  if CursPos > length(St) then
    Beep
  else
    begin
      delete(St,CursPos,1);
      DisplayText
    end
end;

procedure MoveLeft;

begin
  if CursPos = 1 then
    Beep
  else
    dec(CursPos)
end;

procedure MoveRight;

begin
  if CursPos > length(St) then
    Beep
  else
    inc(CursPos)
end;

procedure MoveHome;

begin
  CursPos := 1
end;

procedure MoveEnd;

begin
  CursPos := length(St) + 1
end;

procedure DeleteLine;

begin
  CursPos := 1;
  St := '';
  DisplayText
end;

procedure InsertChar;

begin
  with InputParam do
    if length(St) = MaxLen then
      Beep
    else
      begin
        insert(K,St,CursPos);
        inc(CursPos);
        DisplayText
      end
end;

procedure OvertypeChar;

begin
  if (length(St) = InputParam.MaxLen) and (CursPos > length(St)) then
    Beep
  else
    begin
      if CursPos > length(St) then
        inc(St[0]);
      St[CursPos] := K[1];
      inc(CursPos);
      DisplayText
    end
end;

begin { InputString }
  OldX := WhereX;
  OldY := WhereY;
  OldC := TextAttr;
  OldWMin := WindMin;
  OldWMax := WindMax;
  InitEdit;
  while Editing do
    begin
      GotoXY(CursPos,1);
      repeat
        MBPressed := false;
        K := ScanKey;
        if RightButtonPressCount(MX,MY) > 0 then
          begin
            MBPressed := true;
            Result := inp_RightButtonPressed
          end
        else if LeftButtonPressCount(MX,MY) > 0 then
          begin
            MBPressed := true;
            Result := inp_LeftButtonPressed
          end
      until (K <> '') or MBPressed;
      if MBPressed then
        MouseAbort
      else
        begin
          ConvertWSKey(K);
          if length(K) = 1 then
            begin
              if InputParam.MakeUpper then
                K := upcase(K[1]);
              if K < ' ' then
                case K[1] of
                  ^H  : DeleteLeft;  { <BS> }
                  ^M  : begin
                          Editing := false;
                          Result := inp_Normal;
                          InputParam.NewEdit := true
                        end
                else { Control key - Exit editing }
                  AbortEditing
                end
              else { ASCII character - insert/overtype it, if allowed }
                if K[1] in InputParam.DataSet then
                  begin
                    if InsMode then
                      InsertChar
                    else
                      OvertypeChar
                  end
                else
                  Beep
            end
          else { Function key }
            case K[2] of
              #075 : MoveLeft;               { <Left Arrow > }
              #077 : MoveRight;              { <Right Arrow > }
              #082 : begin                   { <INS> }
                       InsMode := not InsMode;
                       SetCursor
                     end;
              #083 : DeleteChar;             { <DEL> }
              #071 : MoveHome;               { <HOME> }
              #079 : MoveEnd;                { <END> }
              #163 : DeleteLine;             { <ALT DEL> or ^Y }
            else
              AbortEditing
            end
        end
    end;
  Window(lo(OldWMin)+1,hi(OldWMin)+1,lo(OldWMax)+1,hi(OldWMax)+1);
  GotoXY(OldX,OldY);
  TextAttr := OldC;
  RestorePointer
end;

procedure InputPassword (var St : string; var InputParam : InputRec;
                         var Result : InputResultCode);

var
  OldX, OldY,
  OldC         : integer;
  OldWMin,
  OldWMax      : word;
  Editing      : boolean;
  K            : KeyStr;
  CCount       : longint;
  CDisp        : boolean;

begin { InputPassword }
  OldX := WhereX;
  OldY := WhereY;
  OldC := TextAttr;
  Editing := true;
  CDisp := false;
  CCount := 0;
  St := '';
  with InputParam do
  GotoXY(X,Y);
  while Editing do
    begin
      K := InKey;
      if length(K) = 1 then
        begin
          if InputParam.MakeUpper then
            K := upcase(K[1]);
          if K < ' ' then
            if K = ^M then
              begin
                Editing := false;
                Result := inp_Normal
              end
            else { Control character }
              Beep
          else { ASCII character }
            if length(St) = InputParam.MaxLen then
              Beep
            else
              St := St + K
        end
      else { Function key }
        Beep
    end;
  GotoXY(OldX,OldY);
  TextAttr := OldC;
  RestorePointer
end;

end.
