program Trader;

{**************************************************************************}
{*                                                                        *}
{*                             Star Traders                               *}
{*                                                                        *}
{*                     (C) J.N.Zaitseff, 1990-1994.                       *}
{*                                                                        *}
{**************************************************************************}

{ Star Traders is a game of interstellar trading, where the object of the
  game is to create companies, buy and sell shares, borrow and repay money,
  in order to become the wealthiest player (the winner).

  Star Traders is executed thus:

      TRADER <return>                     - normal game.
      TRADER n <return>                   - continue game "n" (1-9).

  History:

  1990:   MicroBee disk version written (under TP 3.01)
  1992:   NEC APC under CP/M-86 version; colour added (under TP 2.00)
  1993:   IBM compatible version, with ANSI.SYS (TP 7.0)
  1994:   IBM compatible version, recoded for colour and windows look (BP 7.0)

  This version of Star Traders is customised for IBM compatibles running
  MS-DOS. It has NOT been rewritten since 1990, only customised.

  All program versions copyright by J.N.Zaitseff, 1990-1994. }

{$IFNDEF MSDOS}
{$IFNDEF DPMI}
    Error: MS-DOS / DPMI compilation only.
{$ENDIF}
{$ENDIF}

{ $DEFINE DEBUG}

{$A+,B-,E+,F-,O-,N+}
{$M 16384,21000,655360}

{$IFDEF DEBUG} {$D+,I+,L+,Q+,R+,S+,Y+} {$ELSE} {$D-,I-,Q-,R-,S-} {$ENDIF}

{$IFDEF DPMI} {$G+} {$ELSE} {$G-} {$ENDIF}

{ $DEFINE TODISK}   { Send windows to disk if defined }

uses
  Crt, Dos, Keyb, TxWindows, TxInput;

const
  Version              = '5.4';
  MaxX                 = 38;      { Map dimensions X x Y }
  MaxY                 = 12;
  StarRatio            = 0.10;    { 10% of map should be stars }
  NumberMoves          = 20;      { Number of moves on the map per turn }
  DefaultMaxTurn       = 50;      { Default number of turns }
  MaxPlayers           = 8;       { Maximum number of players, etc }
  InitialCash          = 6000.0;  { Player's initial cash }
  MaxOverdraft         = 1000.00; { Maximum value player can go negative }
  ProbBankruptcy       = 0.07;    { If overdraft, 7% chance of bankruptcy }
  MaxCompanies         = 8;
  InitialSharePrice    = 60.00;
  SharePriceInc        = 60.00;   { Share price increment for inc shipping }
  SharePriceIncOutpost = 70.00;   { Increment for adding an outpost }
  SharePriceIncStar    = 300.00;  { Increment for adding a star }
  IncSharePrice        = 0.30;    { 30% of incrementing randomly a share price }
  DecSharePrice        = 0.65;    { 65% x 30% chance of decrementing same }
  InitialReturn        = 0.05;    { Initial share return: 5% }
  ProbIncReturn        = 0.25;    { When expanding company, does return inc? }
  IncCompanyReturn     = 0.40;    { Chance of randomly incrementing return }
  MaxCompanyReturn     = 0.40;    { Maximum company return is 40% of share price }
  CompanyBankruptcy    = 0.01;    { 1% chance of company bankruptcy }
  AllAssetsTaken       = 0.20;    { 20% chance of company's assets taken }
  InitialInterestRate  = 0.10;    { 10% }
  IncInterestRate      = 0.30;    { 30% chance of incrementing interest rate }
  MaxInterestRate      = 0.30;    { Maximum interest rate is 30% }
  BidChance            = 0.75;    { 75% chance of successful bidding }
  MaxSharesBidded      = 200;     { Maximum number of shares issued }

  Esc = ^[;

type
  float      = double;        { Floating point type }
  str_       = string[40];
  CompanyRec = record
                 Name         : str_;
                 SharePrice   : float;
                 Return       : float;
                 StockIssued  : integer;
                 MaxStock     : integer;
                 OnMap        : boolean
               end;
  PlayerRec  = record
                 Name         : str_;
                 Cash         : float;
                 Debt         : float;
                 StockOwned   : array [1..MaxCompanies] of integer;
                 InGame       : boolean
               end;
  MapVal     = (Empty,Outpost,Star,C_A,C_B,C_C,C_D,C_E,C_F,C_G,C_H);
  MoveRec    = record
                 X, Y : integer
               end;

const
  CompanyVal : set of MapVal = [C_A..C_H];

var
  Company       : array [1..MaxCompanies] of CompanyRec;
  Player        : array [1..MaxPlayers] of PlayerRec;
  Map           : array [1..MaxX,1..MaxY] of MapVal;
  Move          : array [1..NumberMoves] of MoveRec;
  Selection     : char;
  InterestRate  : float;
  QuitSelected  : boolean;
  TurnNumber    : integer;
  MaxTurn       : integer;
  CurrentPlayer : integer;
  NumberPlayers : integer;
  FirstPlayer   : integer;
  GameLoaded    : boolean;
  GameNumber    : integer;


{ Utility routines }

function UKey : char;                { Get an upper-cased key }

var
  K : KeyStr;

begin
  repeat
    K := InKey
  until length(K) = 1;
  UKey := upcase(K[1])
end;

procedure SetCursorOn (On : boolean); assembler;   { Turn cursor on or off }

asm
        int     11h             { Get equipment list }
        mov     cl,4
        shr     ax,cl
        and     al,3
        mov     cx,2007h
        cmp     al,3            { CGA mode? }
        jnz     @1              { No:  7 scan lines for cursor }
        mov     cl,13           { Yes: 13 scan lines for cursor }
@1:     cmp     On,true
        jnz     @2
        mov     ch,cl
        dec     ch
@2:     mov     ah,1            { Set cursor }
        int     10h
end;

function Answer : char;        { Input "Y" or "N", display "Yes" or "No". }

var
  C : char;

begin
  SetCursorOn(true);
  C := Keyb.Answer;
  if C = 'Y' then
    writeln('Yes.')
  else
    writeln('No.');
  Answer := C;
  SetCursorOn(false)
end;

function InputString (sX, sY : integer) : str_;
{ Input a string. The following characters are allowed: "A" to "Z", "a" to
  "z", "0" to "9" and " ". Maximum length = 40. Blank lines are not allowed }

var
  S          : string;
  InputParam : InputRec;
  Result     : InputResultCode;

begin
  with InputParam do
    begin
      DataSet := Alphanumerics;
      MakeUpper := false;
      Default := '';
      UseDefault := true;
      NewEdit := true;
      MaxLen := 40;
      FCol := White;
      BCol := Black;
      X := sX;
      Y := sY
    end;
  SetCursorOn(true);
  repeat
    TxInput.InputString(S,InputParam,Result)
  until (Result = inp_Normal) and (S <> '');
  SetCursorOn(false);
  InputString := S
end;

function InputNumber (sX, sY : integer) : float;
{ Input a positive real number. Same editing as above. }

var
  S          : string;
  InputParam : InputRec;
  Result     : InputResultCode;
  Temp       : float;
  I          : integer;

begin
  with InputParam do
    begin
      DataSet := ['0'..'9','.'];
      MakeUpper := false;
      Default := '';
      UseDefault := true;
      NewEdit := true;
      MaxLen := 18;
      FCol := White;
      BCol := Black;
      X := sX;
      Y := sY
    end;
  SetCursorOn(true);
  repeat
    TxInput.InputString(S,InputParam,Result)
  until Result = inp_Normal;
  SetCursorOn(false);
  if S = '' then
    S := '0';
  val(S,Temp,I);
  InputNumber := Temp
end;

procedure Title;             { Display the title to screen }

begin
  OpenWindow(1,1,80,4, Yellow,Blue,White, SingleLineBorder);
  TextBackground(Black);
  CentreText('      Star Traders      ',1);
  SetDefaultColours;
  CentreText('By J.N.Zaitseff, 1990.',2)
end;

procedure Continue;          { Wait for the player to press any key }

var
  C : char;

begin
  TextColor(LightGray);
  CentreText('[ Press <SPACE> to continue ]',WinHeight);
  WaitForKey
end;

{ Actual game procedures and functions start here. }

procedure Init;           { Global initialisation }

begin
  SetCursorOn(false);
{$IFDEF TODISK}
  SetWindowStorageToDisk;
{$ENDIF}
  InitTxWindows;
  DrawDefaultBackground;
  Title
end;

procedure DeInit;

begin
  CloseWindow; { Title }
  SetCursorOn(true)
end;

procedure InitData;       { Initialise company data and map }

var
  X, Y : integer;

begin
  QuitSelected := false;
  randomize;
  Company[1].Name := 'Altair Starways';
  Company[2].Name := 'Betelgeuse, Ltd';
  Company[3].Name := 'Capella Freight Co';
  Company[4].Name := 'Denebola Shippers';
  Company[5].Name := 'Eridani Expediters';
  Company[6].Name := 'Fornax Express';
  Company[7].Name := 'Gemeni Inc';
  Company[8].Name := 'Hercules & Co';
  if not GameLoaded then
    begin
      for X := 1 to MaxCompanies do
        with Company[X] do
          begin
            SharePrice := 0.00;
            Return := InitialReturn;
            StockIssued := 0;
            MaxStock := 0;
            OnMap := false
          end;
      InterestRate := InitialInterestRate;
      for X := 1 to MaxX do
        for Y := 1 to MaxY do
          if random < StarRatio then
            Map[X,Y] := Star
          else
            Map[X,Y] := Empty
    end
end;

procedure Instructions;    { Display instructions to screen }

var
  FName : PathStr;
  DStr  : DirStr;
  NStr  : NameStr;
  EStr  : ExtStr;
  C     : char;
  I     : integer;
  F     : text;

begin
  FName := ParamStr(0);
  FSplit(FName,DStr,NStr,EStr);
  FName := DStr + NStr + '.HTX';
  assign(F,FName);
{$I-}
  reset(F);
{$IFDEF DEBUG} {$I+} {$ENDIF}
  if IOResult <> 0 then
    begin
      if length(FName) < 30 then
        I := 56
      else
        I := length(FName) + 26;
      OpenWindow((80-I) div 2,9,I,7, White,Red,Yellow, SingleLineBorder);
      TextBackground(Black);
      CentreText('  Help not available  ',1);
      SetDefaultColours;
      CentreText('Overlay file "'+FName+'" not found.',3);
      Continue;
      CloseWindow
    end
  else
    begin
      OpenWindow(1,6,80,20, Yellow,Blue,White, DoubleLineBorder);
      while not eof(F) do
        begin
          read(F,C);
          case C of
            ^F  : begin
                    Continue;
                    SetDefaultColours;
                    ClrScr
                  end;
            '^' : begin
                    read(F,C);
                    case C of
                      '^' : write(C);
                      'A' : SetDefaultColours;
                      'B' : TextBackground(Black);
                      'C' : TextColor(White);
                      'D' : TextColor(LightBlue);
                      'E' : TextColor(LightGreen);
                      'F' : TextColor(Yellow);
                      'G' : TextColor(LightRed);
                      'H' : begin
                              TextColor(LightBlue);
                              TextBackground(Black)
                            end;
                      'I' : begin
                              TextColor(White);
                              TextBackground(Red)
                            end
                    end
                  end;
            '~' : begin
                    read(F,C);
                    case C of
                      '~' : write(C);
                      'x' : write(MaxX:2);
                      'y' : write(MaxY:2);
                      'm' : write(NumberMoves:2);
                      'M' : write(chr(NumberMoves+$60));
                      'c' : write(MaxCompanies:2);
                      't' : write(MaxTurn:3)
                    end
                   end
          else
            write(C)
          end
        end;
      close(F);
      Continue;
      CloseWindow
    end
end;

procedure InitPlayers;     { Initialise player data or load a game from disk }

var
  I, J : integer;
  C    : char;
  S    : str_;

procedure LoadGame (GN : integer);   { Load a game from disk }

var
  F  : text;
  S  : str_;
  I,
  J,
  K  : integer;
  Gm : char;

begin
  Gm := chr(GN+$30);
  S := 'TRADER'+Gm+'.GME';
  assign(F,S);
{$I-}
  reset(F);
{$IFDEF DEBUG} {$I+} {$ENDIF}
  if IOResult <> 0 then
    begin
      OpenWindow(19,13,42,7, White,Red,Yellow, DoubleLineBorder);
      TextBackground(Black);
      CentreText('  Game not found  ',1);
      TextBackground(Red);
      CentreText('Game '+Gm+' has not been saved to disk.',3);
      Continue;
      CloseWindow;
      C := ' '
    end
  else
    begin
      readln(F,S);
      readln(F,S);
      if S <> Version then
        begin
          close(F);
          OpenWindow(11,13,58,7, White,Red,Yellow, DoubleLineBorder);
          TextBackground(Black);
          CentreText('  Wrong version of Star Traders  ',1);
          TextBackground(Red);
          CentreText('Game '+Gm+' was saved under an old ' +
                     'version of Star Traders.',3);
          Continue;
          CloseWindow;
          C := ' '
        end
      else
        begin
          readln(F,MaxTurn);
          readln(F,NumberPlayers);
          for I := 1 to NumberPlayers do
            with Player[I] do
              begin
                readln(F,Name);
                readln(F,Cash);
                readln(F,Debt);
                readln(F,J);
                if J = 0 then
                  InGame := true
                else
                  InGame := false;
                for J := 1 to MaxCompanies do
                  readln(F,StockOwned[J])
              end;
          for I := 1 to MaxCompanies do
            with Company[I] do
              begin
                readln(F,SharePrice);
                readln(F,Return);
                readln(F,StockIssued);
                readln(F,MaxStock);
                readln(F,J);
                if J = 0 then
                  OnMap := true
                else
                  OnMap := false
              end;
          readln(F,InterestRate);
          readln(F,TurnNumber);
          readln(F,CurrentPlayer);
          readln(F,FirstPlayer);
          for J := 1 to MaxY do
            for I := 1 to MaxX do
              begin
                repeat
                  read(F,C)
                until not (C in [^M,^J]);
                Map[I,J] := MapVal(ord(C)-$40)
              end;
          close(F);
          C := 'C';
          QuitSelected := false;
          GameNumber := GN;
          GameLoaded := true
        end
    end
end;

begin { InitPlayers }
  GameLoaded := false;
  GameNumber := 0;
  if ParamCount > 0 then
    if ((ParamStr(1) <= '9') and (ParamStr(1) >= '1')) then
      begin
        S := ParamStr(1);
        C := S[1];
        OpenWindow(27,10,25,5, Black,LightGray,White, DoubleLineBorder);
        CentreText('Loading game '+C+'...',2);
        LoadGame(ord(C)-$30);
        CloseWindow
      end;
  if not GameLoaded then
    begin
      OpenWindow(9,7,63,5, Yellow,Blue,White, DoubleLineBorder);
      repeat
        ClrScr;
        GotoXY(1,2);
        NumberPlayers := 0;
        write('Enter number of players [');
        TextBackground(Black);
        write('1');
        TextBackground(Blue);
        write('-');
        TextBackground(Black);
        write(MaxPlayers);
        TextBackground(Blue);
        write('] or ');
        TextBackground(Black);
        write('<C>');
        TextBackground(Blue);
        write(' to continue a game: ');
        SetCursorOn(true);
        repeat
          C := UKey;
        until (C = 'C') or ((C >= '1') and (C <= chr(MaxPlayers+$30)));
        if C <> 'C' then
          NumberPlayers := ord(C)-$30;
        SetCursorOn(false);
        write(C);
        if C = 'C' then
          begin
            OpenWindow(15,10,50,5, Yellow,Blue,White, DoubleLineBorder);
            GotoXY(1,2);
            write('Enter game number [');
            TextBackground(Black);
            write('1');
            TextBackground(Blue);
            write('-');
            TextBackground(Black);
            write('9');
            TextBackground(Blue);
            write('] or ');
            TextBackground(Black);
            write('<ESC>');
            TextBackground(Blue);
            write(' to cancel: ');
            SetCursorOn(true);
            repeat
              C := UKey
            until C in ['1'..'9',Esc];
            SetCursorOn(false);
            if C <> Esc then
              begin
                write(C);
                OpenWindow(27,13,25,5, Black,LightGray,White, DoubleLineBorder);
                CentreText('Loading game '+C+'...',2);
                LoadGame(ord(C)-$30);
                CloseWindow
              end;
            CloseWindow
          end
      until (C = 'C') or (NumberPlayers in [1..MaxPlayers]);
      if C <> 'C' then
        begin
          MaxTurn := DefaultMaxTurn;
          if ParamStr(1) = 'MAXTURN' then
            begin
              val(ParamStr(2),I,J);
              if (J = 0) and (I >= 10) and (I <= 500) then
                MaxTurn := I
            end;
          OpenWindow(2,13,76,NumberPlayers+4, Yellow,Blue,White, DoubleLineBorder);
          for I := 1 to NumberPlayers do
            begin
              writeln;
              TextBackground(Black);
              write(' Player ',I,' ');
              TextBackground(Blue);
              write(' Enter your name: ');
              Player[I].Name := InputString(29,I+1);
              SetDefaultColours;
              GotoXY(29,I+1);
              write('':41);
              GotoXY(29,I+1);
              TextColor(White);
              write(Player[I].Name);
              TextColor(Yellow);
              Player[I].Cash := InitialCash;
              Player[I].Debt := 0.00;
              Player[I].InGame := true;
              for J := 1 to MaxCompanies do
                Player[I].StockOwned[J] := 0
            end;
          OpenWindow(15,10,50,5, Yellow,Blue,White, DoubleLineBorder);
          writeln;
          if NumberPlayers = 1 then
            write('  Do you need any ')
          else
            write('Does any player need ');
          write('instructions? [');
          TextBackground(Black);
          write('Y');
          SetDefaultColours;
          write('/');
          TextBackground(Black);
          write('N');
          SetDefaultColours;
          write('] ');
          C := Answer;
          if C = 'Y' then
            Instructions;
          CloseWindow;          { Do you need instructions? }
          CloseWindow;          { Enter names               }
          CloseWindow;          { Enter number of players   }

          OpenWindow(21,11,37,6, Yellow,Blue,White, DoubleLineBorder);
          writeln;
          write('  The game ends after ');
          TextBackground(Black);
          write(MaxTurn);
          TextBackground(Blue);
          write(' moves.');
          Continue;
          TurnNumber := 1;
          CurrentPlayer := 1;
          FirstPlayer := 1;
          if NumberPlayers <> 1 then
            begin
              FirstPlayer := random(NumberPlayers)+1;
              CurrentPlayer := FirstPlayer;
              OpenWindow(17,15,45,7, Yellow,Blue,White, DoubleLineBorder);
              CentreText('The first player to go is',2);
              TextBackground(Black);
              CentreText(' '+Player[FirstPlayer].Name+' ',3);
              TextBackground(Blue);
              Continue;
              CloseWindow
            end
        end;
      CloseWindow
    end
end;

procedure NextPlayer;        { Get the next player }

var
  AllOut : boolean;
  I      : integer;

begin
  AllOut := true;
  for I := 1 to NumberPlayers do
    if Player[I].InGame then
      AllOut := false;
  if not AllOut then
    repeat
      CurrentPlayer := CurrentPlayer + 1;
      if CurrentPlayer = NumberPlayers + 1 then
        CurrentPlayer := 1;
      if CurrentPlayer = FirstPlayer then
        TurnNumber := TurnNumber + 1
    until Player[CurrentPlayer].InGame
end;

procedure SelectMoves;       { Select moves "a" to "MaxTurn" }

procedure SortMoves;         { Sort moves from left to right }

var
  I, J, K : integer;
  Temp    : MoveRec;

begin
  for I := 1 to NumberMoves - 1 do
    begin
      K := I;
      Temp := Move[I];
      for J := I+1 to NumberMoves do
        if Move[J].X < Temp.X then
          begin
            K := J;
            Temp := Move[J]
          end;
      Move[K] := Move[I];
      Move[I] := Temp
    end
end;

var
  I, J, TempX, TempY : integer;
  OK                 : boolean;

begin { SelectMoves }
  I := 0;
  for TempX := 1 to MaxX do
    for TempY := 1 to MaxY do
      if Map[TempX,TempY] = Empty then
        I := I + 1;
  if I < NumberMoves then
    QuitSelected := true
  else
    begin
      for I := 1 to NumberMoves do
        begin
          repeat
            TempX := random(MaxX)+1;
            TempY := random(MaxY)+1;
            OK := true;
            if I > 1 then
              for J := I-1 downto 1 do
                if ((Move[J].X = TempX) and (Move[J].Y = TempY)) then
                  OK := false
          until OK and (Map[TempX,TempY] = Empty);
          Move[I].X := TempX;
          Move[I].Y := TempY
        end;
      QuitSelected := false;
      SortMoves
    end
end;

procedure DisplayMap;            { Display the map to screen in colour }

var
  X, Y, I : integer;

begin
  OpenWindow(1,5,80,16, White,Black,LightGray, SingleLineBorder);
  TextBackground(Blue);
  TextColor(Yellow);
  ClrEOL;
  write('  Current Player: ');
  TextColor(White);
  write(Player[CurrentPlayer].Name);
  TextColor(Yellow);
  write('. Turn #');
  TextColor(White);
  write(TurnNumber);
  if TurnNumber = MaxTurn then
    begin
      TextColor(Blink+White);
      GotoXY(59,WhereY);
      write('*** Last turn ***')
    end;
  TextBackground(Black);
  TextColor(LightGray);
  Window(1,7,80,20);
  write('');
  for X := 1 to 78 do
    write('');
  write('');
  SetDefaultColours;
  SetWindowViewPort;
  writeln;
  writeln;
  for Y := 1 to MaxY do
    begin
      for X := 1 to MaxX do
        begin
          if (MaxX < 40) and (X > 1) then
            write(' ');
          case Map[X,Y] of
            Empty   : begin
                        TextColor(LightBlue);
                        write('.')
                      end;
            Outpost : begin
                        TextColor(LightGreen);
                        write(#4)
                      end;
            Star    : begin
                        TextColor(Yellow);
                        write('*')
                      end;
          else
            TextColor(LightRed);
            write(chr(ord(Map[X,Y])-ord(C_A)+$41))
          end
        end;
      if Y < MaxY then
        writeln
    end
end;

function TotalValue (N : integer) : float;       { Calculate the total worth
                                                  of player N }

var
  I    : integer;
  Temp : float;

begin
  Temp := Player[N].Cash - Player[N].Debt;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      Temp := Temp + Player[N].StockOwned[I]*Company[I].SharePrice;
  TotalValue := Temp
end;

procedure DisplayStatus;            { Display the current player's status }

var
  Temp      : float;
  I         : integer;
  AllOffMap : boolean;

begin
  OpenWindow(1,3,80,23, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Stock Owned  ',1);
  TextBackground(Blue);
  GotoXY((WinWidth-length(Player[CurrentPlayer].Name)-8) div 2,2);
  write('Player: ');
  TextColor(White);
  write(Player[CurrentPlayer].Name);
  TextColor(Yellow);
  writeln('.');
  if TotalValue(CurrentPlayer) = 0 then
    begin
      TextColor(LightGray);
      CentreText('[ Press <SPACE> to continue ]',WinHeight);
      OpenWindow(21,12,40,5, Blink+White,Red,Yellow, DoubleLineBorder);
      CentreText('* * *   B A N K R U P T   * * *',2);
      WaitForKey;
      CloseWindow;
      CloseWindow       { Stock Owned }
    end
  else
    begin
      AllOffMap := true;
      for I := 1 to MaxCompanies do
        if Company[I].OnMap then
          AllOffMap := false;
      if AllOffMap then
        CentreText('No companies on the map.',9)
      else
        begin
          GotoXY(1,4);
          TextBackground(Black);
          TextColor(White);
          ClrEOL;
          writeln('                          Price per               Holdings   Ownership of');
          ClrEOL;
          writeln('  Company                 share ($)   Return (%)  (shares)    company (%)');
          TextBackground(Blue);
          for I := 1 to MaxCompanies do
            if Company[I].OnMap then
              begin
                TextColor(White);
                write('  ',Company[I].Name);
                TextColor(Yellow);
                GotoXY(25,WhereY);
                write(Company[I].SharePrice:10:2,'   ');
                write(Company[I].Return*100:8:2,'   ');
                write(Player[CurrentPlayer].StockOwned[I]:8,'  ');
                if Company[I].StockIssued = 0 then
                  Temp := 0.00
                else
                  Temp := Player[CurrentPlayer].StockOwned[I]
                          / Company[I].StockIssued;
                writeln(Temp*100:12:2)
              end
        end;
      GotoXY(16,15);
      write('Current cash:       ');
      TextBackground(Black);
      writeln(' $',Player[CurrentPlayer].Cash:4:2,' ');
      TextBackground(Blue);
      if Player[CurrentPlayer].Debt <> 0.00 then
        begin
          write('':15,'Current debt:       ');
          TextBackground(Black);
          writeln(' $',Player[CurrentPlayer].Debt:4:2,' ');
          TextBackground(Blue);
          write('':15,'Debt interest rate: ');
          TextBackground(Black);
          writeln(' ',InterestRate*100:4:2,'% ');
          TextBackground(Blue)
        end;
      writeln;
      TextColor(White);
      write('':15,'Total value:        ');
      TextBackground(Black);
      TextColor(Yellow);
      writeln(' $',TotalValue(CurrentPlayer):4:2,' ');
      TextBackground(Blue);
      Continue;
      CloseWindow
    end
end;

procedure GetMove;            { Wait for player to enter their move }

var
  I, X, Y : integer;
  C       : char;

begin
  if QuitSelected then
    Selection := Esc
  else
    begin
      DisplayMap;
      TextColor(White);
      TextBackground(Red);
      for I := 1 to NumberMoves do
        begin
          X := Move[I].X;
          Y := Move[I].Y;
          GotoXY(X*2-1,Y+2);
          write(chr(I+$60));
        end;
      OpenWindow(1,21,80,5, Yellow,Blue,White, DoubleLineBorder);
      repeat
        repeat
          ClrScr;
          GotoXY(1,2);
          TextBackground(Black);
          write('<1>');
          TextBackground(Blue);
          write(' Display stock owned.');
          GotoXY(40,2);
          TextBackground(Black);
          write('<2>');
          TextBackground(Blue);
          write('  Declare bankruptcy.');
          GotoXY(1,3);
          TextBackground(Black);
          write('<3>');
          TextBackground(Blue);
          write(' Save game.');
          GotoXY(39,3);
          TextBackground(Black);
          write('<ESC>');
          TextBackground(Blue);
          write(' Quit the game.');

          GotoXY(11,1);
          write('Select move [');
          TextColor(White);
          TextBackground(Red);
          write('a');
          SetDefaultColours;
          write('-');
          TextColor(White);
          TextBackground(Red);
          write(chr(NumberMoves+$60));
          SetDefaultColours;
          write('/');
          TextBackground(Black);
          write('1');
          TextBackground(Blue);
          write('-');
          TextBackground(Black);
          write('3');
          TextBackground(Blue);
          write('/');
          TextBackground(Black);
          write('<ESC>');
          TextBackground(Blue);
          write(']: ');
          SetCursorOn(true);
          repeat
            C := UKey;
            if C in ['1'..'3',Esc] then
              I := 1
            else if (C >= 'A') and (C <= chr(NumberMoves+$40)) then
              I := ord(C)-$40
            else
              I := 0
          until I in [1..NumberMoves];
          SetCursorOn(false);
          TextColor(White);
          if C in ['1'..'3',Esc] then
            begin
              if C = Esc then
                write('<ESC> ')
              else
                write('<',C,'> ');
              TextColor(LightGray)
            end;
          case C of
            '1' : writeln('(Display stock owned)');
            '2' : writeln('(Declare bankruptcy)');
            '3' : writeln('(Save game)');
            Esc : writeln('(Quit the game)')
          else
            write('Move ');
            TextColor(White);
            TextBackground(Red);
            write(chr(ord(C)+$20));
            TextBackground(Blue);
            writeln('.')
          end;
          TextColor(Yellow);
          ClrEOL;
          writeln;
          ClrEOL;
          Selection := C;
          if Selection = '1' then
            DisplayStatus
          else
            begin
              GotoXY(20,2);
              write('Are you sure? [');
              TextBackground(Black);
              write('Y');
              TextBackground(Blue);
              write('/');
              TextBackground(Black);
              write('N');
              TextBackground(Blue);
              write('] ');
              TextColor(White);
              C := Answer;
              TextColor(Yellow)
            end
        until C in ['Y','1']
      until C <> '1'
    end
end;

procedure ProcessMove;        { Process the move the player selected }

var
  Left, Right, Up, Down, Temp : MapVal;
  X, Y, I                     : integer;
  AllOut                      : boolean;

procedure AssignSurroundVals;         { What surrounds the X,Y position? }

begin
  if X = 1    then Left  := Empty else Left  := Map[X-1,Y];
  if X = MaxX then Right := Empty else Right := Map[X+1,Y];
  if Y = 1    then Up    := Empty else Up    := Map[X,Y-1];
  if Y = MaxY then Down  := Empty else Down  := Map[X,Y+1]
end;

procedure IncreaseSharePrice (N : integer; P : float);    { Increase the share
                                                           price and return }
var
  I : integer;

begin
  I := round(P);
  Company[N].SharePrice := Company[N].SharePrice + P + (P * random / 2);
  Company[N].MaxStock := Company[N].MaxStock + (I div (random(10)+5));
  if random < ProbIncReturn then
    Company[N].Return := Company[N].Return * (random + 0.60)
end;

procedure IncludeOutpost (Xn, Yn : integer);     { Include any outposts into
                                                   company. (Recursive) }
var
  L, R, U, D : MapVal;

begin
  Map[Xn,Yn] := Temp;
  IncreaseSharePrice(I,SharePriceIncOutpost);
  if Xn = 1    then L := Empty else L := Map[Xn-1,Yn];
  if Xn = MaxX then R := Empty else R := Map[Xn+1,Yn];
  if Yn = 1    then U := Empty else U := Map[Xn,Yn-1];
  if Yn = MaxY then D := Empty else D := Map[Xn,Yn+1];
  if L = Outpost then IncludeOutpost(Xn-1,Yn);
  if R = Outpost then IncludeOutpost(Xn+1,Yn);
  if U = Outpost then IncludeOutpost(Xn,Yn-1);
  if D = Outpost then IncludeOutpost(Xn,Yn+1)
end;

procedure EstablishNewCompany;         { If a position next to an outpost or
                                         star is selected }
var
  I, J    : integer;
  AllUsed : boolean;
  Temp    : MapVal;

const
  CompanyFormers : set of MapVal = [Outpost,Star];

begin
  if (Left in CompanyFormers) or (Right in CompanyFormers) or
     (Up   in CompanyFormers) or (Down  in CompanyFormers) then
    begin
      AllUsed := true;
      for I := MaxCompanies downto 1 do
        if not Company[I].OnMap then
          begin
            AllUsed := false;
            J := I
          end;
      if AllUsed then
        Map[X,Y] := Outpost
      else
        begin
          OpenWindow(20,10,40,8, Yellow,Blue,White, DoubleLineBorder);
          TextBackground(Black);
          CentreText('  New Company  ',1);
          TextBackground(Blue);
          CentreText('A new company has been formed.',3);
          GotoXY((WinWidth-length(Company[J].Name)-11) div 2,4);
          write('Its name is ');
          TextColor(White);
          write(Company[J].Name);
          TextColor(Yellow);
          write('.');
          Continue;
          CloseWindow;
          Temp := MapVal(J+ord(C_A)-1);
          Map[X,Y] := Temp;
          Company[J].Return := InitialReturn;
          Company[J].SharePrice := InitialSharePrice;
          Company[J].OnMap := true;
          Company[J].StockIssued := 5;
          Company[J].MaxStock := 50;
          for I := 1 to NumberPlayers do
            Player[I].StockOwned[J] := 0;
          Player[CurrentPlayer].StockOwned[J] := 5
        end
    end
end;

procedure MergeCompanies (Company1, Company2 : MapVal);   { Perform a merge }

var
  I, J, X1, Y1      : integer;
  Bonus, Val1, Val2 : float;
  ExtraStock        : integer;
  SwapTemp          : MapVal;

begin
  I := ord(Company1)-ord(C_A)+1;
  J := ord(Company2)-ord(C_A)+1;
  Val1 := Company[I].SharePrice * Company[I].StockIssued * Company[I].Return;
  Val2 := Company[J].SharePrice * Company[J].StockIssued * Company[J].Return;
  if Val1 < Val2 then
    begin
      SwapTemp := Company1;
      Company1 := Company2;
      Company2 := SwapTemp;
      X1 := I;
      I := J;
      J := X1
    end;
  OpenWindow(2,12-NumberPlayers,78,NumberPlayers+14,
             Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Company Merger  ',1);
  TextBackground(Blue);
  TextColor(White);
  GotoXY((52-length(Company[J].Name)-length(Company[I].Name)) div 2,3);
  write(Company[J].Name);
  TextColor(Yellow);
  write(' has just merged into ');
  TextColor(White);
  write(Company[I].Name);
  TextColor(Yellow);
  writeln('.');
  CentreText('Please note the following transactions:',5);
  GotoXY(1,7);
  write('Old stock: ');
  TextBackground(Black);
  write(' ',Company[J].Name,' ');
  TextBackground(Blue);
  GotoXY(1,8);
  write('New stock: ');
  TextBackground(Black);
  write(' ',Company[I].Name,' ');
  TextBackground(Blue);
  GotoXY(1,10);
  TextBackground(Black);
  TextColor(White);
  ClrEOL;
  writeln(' Player','':25,'Old stock   New stock   Total   Bonus ($)');
  SetDefaultColours;
  Y1 := 0;
  for X1 := 1 to NumberPlayers do
    if Player[X1].InGame then
      begin
        TextColor(White);
        write(' ',Player[X1].Name);
        TextColor(Yellow);
        GotoXY(31,WhereY);
        ExtraStock := Player[X1].StockOwned[J] div 2;
        write(Player[X1].StockOwned[J]:9,'   ',ExtraStock:9,' ');
        Player[X1].StockOwned[I] := Player[X1].StockOwned[I] + ExtraStock;
        Y1 := Y1 + ExtraStock;
        write(Player[X1].StockOwned[I]:7,'   ');
        if Company[J].StockIssued = 0 then
          Bonus := 0.00
        else
          Bonus := 10 * ((Player[X1].StockOwned[J] / Company[J].StockIssued)
                   * Company[J].SharePrice);
        writeln(Bonus:11:2);
        Player[X1].Cash := Player[X1].Cash + Bonus;
        Player[X1].StockOwned[J] := 0
      end;
  Continue;
  CloseWindow;
  Map[X,Y] := Company2;
  with Company[I] do
    begin
      StockIssued := StockIssued + Y1;
      SharePrice := SharePrice + Company[J].SharePrice / (random + 1.5);
      MaxStock := MaxStock + Y1
    end;
  with Company[J] do
    begin
      StockIssued := 0;
      MaxStock := 0;
      OnMap := false
    end;
  for X1 := 1 to MaxX do
    for Y1 := 1 to MaxY do
      if Map[X1,Y1] = Company2 then
        Map[X1,Y1] := Company1;
  AssignSurroundVals
end;

procedure SaveGame;           { Save the game to disk }

var
  F : text;
  C : char;
  S : str_;
  I, X, Y : integer;

begin
  Title;
  if GameLoaded then
    C := chr(GameNumber+$30)
  else
    begin
      OpenWindow(19,11,42,6, Yellow,Blue,White, DoubleLineBorder);
      TextBackground(Black);
      CentreText('  Save Game  ',1);
      TextBackground(Blue);
      GotoXY(2,3);
      write('Enter game number to save to [');
      TextBackground(Black);
      write('1');
      TextBackground(Blue);
      write('-');
      TextBackground(Black);
      write('9');
      TextBackground(Blue);
      write(']: ');
      SetCursorOn(true);
      repeat
        C := UKey
      until C in ['1'..'9'];
      SetCursorOn(false);
      TextColor(White);
      write(C);
      TextColor(Yellow)
    end;
  OpenWindow(27,14,25,5, Black,LightGray,White, DoubleLineBorder);
  CentreText('Saving game '+C+'...',2);
  S := 'TRADER'+C+'.GME';
  assign(F,S);
  rewrite(F);
  writeln(F,'; Star Traders  ---  (C) J.N.Zaitseff, 1990-1994.');
  writeln(F,Version);
  writeln(F,MaxTurn);
  writeln(F,NumberPlayers);
  for I := 1 to NumberPlayers do
    with Player[I] do
      begin
        writeln(F,Name);
        writeln(F,Cash);
        writeln(F,Debt);
        if InGame then
          writeln(F,'0')
        else
          writeln(F,'1');
        for X := 1 to MaxCompanies do
          writeln(F,StockOwned[X])
      end;
  for I := 1 to MaxCompanies do
    with Company[I] do
      begin
        writeln(F,SharePrice);
        writeln(F,Return);
        writeln(F,StockIssued);
        writeln(F,MaxStock);
        if OnMap then
          writeln(F,'0')
        else
          writeln(F,'1')
      end;
  writeln(F,InterestRate);
  writeln(F,TurnNumber);
  writeln(F,CurrentPlayer);
  writeln(F,FirstPlayer);
  for Y := 1 to MaxY do
    begin
      for X := 1 to MaxX do
        write(F,chr(ord(Map[X,Y])+$40));
      writeln(F)
    end;
  close(F);
  CloseWindow; { Saving game... }
  CloseWindow; { Which game number? }
  if not GameLoaded then
    CloseWindow;    { Map }
  Selection := Esc
end;

begin { ProcessMove }
  if Selection = '2' then          { Declare bankruptcy }
    begin
      I := length(Player[CurrentPlayer].Name);
      OpenWindow((49-I) div 2,11,I+33,7, White,Red,Yellow, DoubleLineBorder);
      TextBackground(Black);
      CentreText('  Bankruptcy Court  ',1);
      TextBackground(Red);
      GotoXY(2,3);
      TextColor(Yellow);
      write(Player[CurrentPlayer].Name);
      TextColor(White);
      write(' has declared bankruptcy.');
      Continue;
      CloseWindow;
      DisplayStatus;
      Player[CurrentPlayer].InGame := false;
      for I := 1 to MaxCompanies do
        begin
          Company[I].MaxStock := Company[I].MaxStock +
                                 Player[CurrentPlayer].StockOwned[I];
          Company[I].StockIssued := Company[I].StockIssued -
                                    Player[CurrentPlayer].StockOwned[I];
          Player[CurrentPlayer].StockOwned[I] := 0
        end;
      Player[CurrentPlayer].Cash := 0.00;
      Player[CurrentPlayer].Debt := 0.00;
      AllOut := true;
      for I := 1 to NumberPlayers do
        if Player[I].InGame then
          AllOut := false;
      if AllOut then
        Selection := Esc
    end;
  if Selection = '3' then         { Save the game }
    SaveGame;
  if Selection = Esc then         { Quit the game }
    QuitSelected := true
  else
    if Player[CurrentPlayer].InGame then
      begin
        I := ord(Selection) - $40;
        X := Move[I].X;
        Y := Move[I].Y;
        AssignSurroundVals;
        if (Left = Empty) and (Right = Empty) and (Up = Empty) and (Down = Empty) then
          Map[X,Y] := Outpost;
        if (Left in CompanyVal) and (Right in CompanyVal) and (Left <> Right) then
          MergeCompanies(Left,Right);
        if (Left in CompanyVal) and (Up in CompanyVal) and (Left <> Up) then
          MergeCompanies(Left,Up);
        if (Left in CompanyVal) and (Down in CompanyVal) and (Left <> Down) then
          MergeCompanies(Left,Down);
        if (Right in CompanyVal) and (Up in CompanyVal) and (Right <> Up) then
          MergeCompanies(Right,Up);
        if (Right in CompanyVal) and (Down in CompanyVal) and (Right <> Down) then
          MergeCompanies(Right,Down);
        if (Up in CompanyVal) and (Down in CompanyVal) and (Up <> Down) then
          MergeCompanies(Up,Down);
        if not((Left in CompanyVal) or (Right in CompanyVal) or (Up in CompanyVal)
           or (Down in CompanyVal)) then
          EstablishNewCompany;
        Temp := Empty;
        if Left  in CompanyVal then Temp := Left;
        if Right in CompanyVal then Temp := Right;
        if Up    in CompanyVal then Temp := Up;
        if Down  in CompanyVal then Temp := Down;
        if Temp <> Empty then
          begin
            I := ord(Temp)-ord(C_A)+1;
            Map[X,Y] := Temp;
            IncreaseSharePrice(I,SharePriceInc)
          end;
        Temp := Map[X,Y];
        I := 0;
        if Temp in CompanyVal then I := ord(Temp)-ord(C_A)+1;
        if I > 0 then
          begin
            if Left  = Star then IncreaseSharePrice(I,SharePriceIncStar);
            if Right = Star then IncreaseSharePrice(I,SharePriceIncStar);
            if Up    = Star then IncreaseSharePrice(I,SharePriceIncStar);
            if Down  = Star then IncreaseSharePrice(I,SharePriceIncStar);
          end;
        if Temp in CompanyVal then
          begin
            if Left  = Outpost then IncludeOutpost(X-1,Y);
            if Right = Outpost then IncludeOutpost(X+1,Y);
            if Up    = Outpost then IncludeOutpost(X,Y-1);
            if Down  = Outpost then IncludeOutpost(X,Y+1)
          end
      end;
  CloseWindow; { Choose move }
  CloseWindow  { Map         }
end;

procedure StockExchange;      { Enter the Interstellar Stock Exchange }

var
  AllOff  : boolean;
  I, J    : integer;
  C       : char;
  BidUsed : boolean;
  R, R2   : float;

procedure AdjustValues;       { Adjust the cost of shares, their return, the
                                interest rate, etc. }

var
  I, J, X, Y : integer;
  Temp : MapVal;

begin
  if random > (1 - CompanyBankruptcy) then
    begin
      I := random(MaxCompanies)+1;
      if Company[I].OnMap then
        begin
          R := 0;
          if random > AllAssetsTaken then
            R := random;
          OpenWindow(10,10,60,14, White,Red,Yellow, DoubleLineBorder);
          TextBackground(Black);
          CentreText('  Bankruptcy Court  ',1);
          TextBackground(Red);
          GotoXY((28-I) div 2,3);
          TextColor(Yellow);
          write(Company[I].Name);
          TextColor(White);
          write(' has been declared');
          CentreText('bankrupt by the Interstellar Trading Bank.',4);
          if R = 0 then
            CentreText('All assets have been taken to repay outstanding loans.',8)
          else
            begin
              GotoXY(5,6);
              write('The bank has agreed to pay stock holders ');
              TextColor(Yellow);
              write(R*100:4:2,'%');
              TextColor(White);
              CentreText('of the share value on each share owned.',7);
              GotoXY(10,9);
              write('Old share value:       ');
              TextBackground(Black);
              writeln(' $',Company[I].SharePrice:4:2,' ');
              TextBackground(Red);
              write('':9,'Amount paid per share: ');
              TextBackground(Black);
              writeln(' $',Company[I].SharePrice * R:4:2,' ');
              TextBackground(Red);
              {writeln('Player','':25,'Holdings   Value ($)   Amount paid ($)');}
              for J := 1 to NumberPlayers do
                if Player[J].InGame then
                  begin
                    {write(Player[J].Name,'':25,Player[J].StockOwned[I]:8);}
                    R2 := Player[J].StockOwned[I] * Company[I].SharePrice;
                    {writeln('   ',R2:9:2,'   ',R2 * R:15:2);}
                    Player[J].Cash := Player[J].Cash + R2 * R
                  end
            end;
          Continue;
          CloseWindow;
          for J := 1 to NumberPlayers do
            Player[J].StockOwned[I] := 0;
          with Company[I] do
            begin
              SharePrice := 0.00;
              Return := 0.00;
              StockIssued := 0;
              MaxStock := 0;
              OnMap := false
            end;
          Temp := MapVal(I+ord(C_A)-1);
          for X := 1 to MaxX do
            for Y := 1 to MaxY do
              if Map[X,Y] = Temp then
                Map[X,Y] := Empty
        end
    end;
  if random < IncCompanyReturn then
    begin
      I := random(MaxCompanies)+1;
      Company[I].Return := Company[I].Return * (random + 0.75)
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap and (Company[I].Return > MaxCompanyReturn) then
      Company[I].Return := Company[I].Return / (random + 1.5);
  if random < IncSharePrice then
    begin
      I := random(MaxCompanies)+1;
      R := random * (Company[I].SharePrice / 4);
      if random < DecSharePrice then
        R := -R;
      Company[I].SharePrice := Company[I].SharePrice + R
    end;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      begin
        if Company[I].StockIssued = 0 then
          R := 0
        else
          R := Player[CurrentPlayer].StockOwned[I] / Company[I].StockIssued;
        with Player[CurrentPlayer] do
          Cash := Cash + (StockOwned[I] * Company[I].SharePrice *
                  Company[I].Return) + (R * Company[I].SharePrice * 2)
      end;
  if random < IncInterestRate then
    InterestRate := InterestRate * (random + 0.65);
  if InterestRate > MaxInterestRate then
    InterestRate := InterestRate / (random + 1.5);
{  if Player[CurrentPlayer].Debt * InterestRate > 10000.00 then
    R := 0.5
  else
    R := 1; }
  with Player[CurrentPlayer] do
    Debt := Debt * (1 + InterestRate * R)
end;

procedure VisitBank;            { Visit the Interstellar Trading Bank }

var
  R, R2 : float;
  C     : char;

begin
  R2 := 2 * (TotalValue(CurrentPlayer) - Player[CurrentPlayer].Debt);
  if R2 < 0.00 then
    R2 := 0.00;

  OpenWindow(2,10,78,9, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Interstellar Trading Bank  ',1);
  TextBackground(Blue);
  GotoXY(24,3);
  write('Current cash:  ');
  TextBackground(Black);
  write(' $',Player[CurrentPlayer].Cash:4:2,' ');
  TextBackground(Blue);
  GotoXY(24,4);
  write('Current debt:  ');
  TextBackground(Black);
  write(' $',Player[CurrentPlayer].Debt:4:2,' ');
  TextBackground(Blue);
  GotoXY(24,5);
  write('Interest rate: ');
  TextBackground(Black);
  write(' ',InterestRate * 100:4:2,'% ');
  TextBackground(Blue);
  GotoXY(24,6);
  write('Credit limit:  ');
  TextBackground(Black);
  writeln(' $',R2:4:2,' ');
  TextBackground(Blue);
  OpenWindow(2,19,78,7, Yellow,Blue,White, DoubleLineBorder);
  GotoXY(24,3);
  TextBackground(Black);
  write('<1>');
  TextBackground(Blue);
  write(' Borrow money.');
  GotoXY(24,4);
  TextBackground(Black);
  write('<2>');
  TextBackground(Blue);
  write(' Repay debt.');
  GotoXY(24,5);
  TextBackground(Black);
  write('<3>');
  TextBackground(Blue);
  write(' Exit from the Bank.');
  GotoXY(23,1);
  write('Enter selection [');
  TextBackground(Black);
  write('1');
  TextBackground(Blue);
  write('/');
  TextBackground(Black);
  write('2');
  TextBackground(Blue);
  write('/');
  TextBackground(Black);
  write('3');
  TextBackground(Blue);
  write(']: ');
  SetCursorOn(true);
  repeat
    C := UKey
  until C in ['1'..'3',' '];
  SetCursorOn(false);
  write(C);
  case C of
    '1' : if R2 = 0.00 then
            begin
              OpenWindow(17,13,47,7, White,Red,Yellow, DoubleLineBorder);
              TextBackground(Black);
              TextColor(Yellow);
              CentreText('  Insufficient Credit Limit  ',1);
              SetDefaultColours;
              CentreText('The bank will not lend you any more money.',3);
              Continue;
              CloseWindow
            end
          else
            begin
              ClrScr;
              GotoXY(12,3);
              write('How much do you wish to borrow? ');
              repeat
                R := InputNumber(45,3);
              until R <= (R2 + 0.01);
              with Player[CurrentPlayer] do
                begin
                  Cash := Cash + R;
                  Debt := (1+ InterestRate) * (Debt + R)
                end
            end;
    '2' : if Player[CurrentPlayer].Debt = 0.00 then
            begin
              OpenWindow(23,13,35,7, White,Red,Yellow, DoubleLineBorder);
              TextBackground(Black);
              TextColor(Yellow);
              CentreText('  No Debt  ',1);
              SetDefaultColours;
              CentreText('You have no debt to pay off.',3);
              Continue;
              CloseWindow
            end
          else
            begin
              ClrScr;
              GotoXY(12,3);
              write('How much do you wish to repay? ');
              repeat
                R := InputNumber(44,3)
              until (R <= (Player[CurrentPlayer].Cash + 0.01)) and
                    (R <= (Player[CurrentPlayer].Debt + 0.01));
              with Player[CurrentPlayer] do
                begin
                  Cash := Cash - R;
                  Debt := Debt - R;
                  if Cash < 0.01 then
                    Cash := 0.00;
                  if Debt < 0.01 then
                    Debt := 0.00
                end
            end
  end;
  CloseWindow;  { Enter selection }
  CloseWindow   { Bank status     }
end;

procedure Trade (N : integer);  { Buy, sell or bid for shares with company N }

var
  R, R2 : float;
  C     : char;
  I, J  : integer;

procedure TradeTitle;

begin
  OpenWindow(2,10,78,9, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Stock transaction in '+Company[N].Name+'  ',1);
  TextBackground(Blue);
  GotoXY(1,3);
  write('Current cash:     ');
  TextBackground(Black);
  write(' $',Player[CurrentPlayer].Cash:4:2,' ');
  TextBackground(Blue);
  GotoXY(1,4);
  write('Current holdings: ');
  TextBackground(Black);
  write(' ',Player[CurrentPlayer].StockOwned[N],' ');
  TextBackground(Blue);
  GotoXY(1,6);
  write('Each share costs ');
  TextColor(White);
  write('$',Company[N].SharePrice:4:2);
  TextColor(Yellow);
  write('. There are ');
  TextColor(White);
  write(Company[N].MaxStock - Company[N].StockIssued);
  TextColor(Yellow);
  write(' shares left.')
end;

begin { Trade }
  OpenWindow(2,9,78,8, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Stock transaction in '+Company[N].Name+'  ',1);
  TextBackground(Blue);
  GotoXY(1,3);
  with Company[N] do
    begin
      write('Shares issued: ');
      TextBackground(Black);
      writeln(' ',StockIssued,' ');
      TextBackground(Blue);
      write('Shares left:   ');
      TextBackground(Black);
      writeln(' ',MaxStock - StockIssued,' ');
      TextBackground(Blue);
      write('Price:         ');
      TextBackground(Black);
      writeln(' $',SharePrice:4:2,' ');
      TextBackground(Blue);
      write('Return:        ');
      TextBackground(Black);
      write(' ',Return*100:4:2,'% ');
      TextBackground(Blue)
    end;
  GotoXY(37,3);
  write('Current cash:     ');
  TextBackground(Black);
  write(' $',Player[CurrentPlayer].Cash:4:2,' ');
  TextBackground(Blue);
  GotoXY(37,4);
  write('Current holdings: ');
  TextBackground(Black);
  write(' ',Player[CurrentPlayer].StockOwned[N],' ');
  TextBackground(Blue);
  if Company[N].StockIssued = 0 then
    R2 := 0.0
  else
    R2 := Player[CurrentPlayer].StockOwned[N] / Company[N].StockIssued;
  GotoXY(37,5);
  write('Percentage owned: ');
  TextBackground(Black);
  write(' ',R2*100:4:2,'% ');
  TextBackground(Blue);

  OpenWindow(2,17,78,6, Yellow,Blue,White, DoubleLineBorder);
  GotoXY(1,3);
  TextBackground(Black);
  write('<1>');
  TextBackground(Blue);
  write(' Buy stock.');
  GotoXY(40,3);
  TextBackground(Black);
  write('<2>');
  TextBackground(Blue);
  write(' Sell stock.');
  GotoXY(1,4);
  TextBackground(Black);
  write('<3>');
  TextBackground(Blue);
  write(' Bid company to issue more shares.');
  GotoXY(40,4);
  TextBackground(Black);
  write('<4>');
  TextBackground(Blue);
  write(' Exit to Stock Exchange.');

  GotoXY(24,1);
  write('Enter selection: [');
  TextBackground(Black);
  write('1');
  TextBackground(Blue);
  write('-');
  TextBackground(Black);
  write('4');
  TextBackground(Blue);
  write(']: ');
  SetCursorOn(true);
  repeat
    C := UKey
  until C in ['1'..'4',' '];
  SetCursorOn(false);
  CloseWindow;  { Select choice... }
  CloseWindow;  { Trading          }
  case C of
    '1' : begin
            R := Player[CurrentPlayer].Cash / Company[N].SharePrice;
            if R > 32000 then
              R := 32000;
            I := trunc(R);
            with Company[N] do
              if I > (MaxStock - StockIssued) then
                I := MaxStock - StockIssued;
            TradeTitle;
            OpenWindow(2,19,78,5, Yellow,Blue,White, DoubleLineBorder);
            GotoXY(22,1);
            write('You can purchase ');
            TextBackground(Black);
            write(' ',I,' ');
            TextBackground(Blue);
            write(' shares.');
            GotoXY(7,3);
            write('How many shares do you wish to purchase? ');
            repeat
              R := InputNumber(48,3);
              if R > 32000 then
                R := 32000;
              J := trunc(R)
            until (J >= 0) and (J <= I);
            CloseWindow;
            CloseWindow; { TradeTitle }
            with Player[CurrentPlayer] do
              begin
                StockOwned[N] := StockOwned[N] + J;
                Company[N].StockIssued := Company[N].StockIssued + J;
                Cash := Cash - (J * Company[N].SharePrice)
              end
          end;
    '2' : begin
            TradeTitle;
            OpenWindow(2,19,78,5, Yellow,Blue,White, DoubleLineBorder);
            if Player[CurrentPlayer].StockOwned[N] = 0 then
              begin
                CentreText('You do not have any shares to sell.',1);
                Continue
              end
            else
              begin
                GotoXY(9,2);
                write('How many shares do you wish to sell? ');
                repeat
                  R := InputNumber(46,2);
                  if R > 32000 then
                    R := 32000;
                  J := trunc(R)
                until (J >= 0) and (J <= Player[CurrentPlayer].StockOwned[N]);
                with Player[CurrentPlayer] do
                  begin
                    StockOwned[N] := StockOwned[N] - J;
                    Company[N].StockIssued := Company[N].StockIssued - J;
                    Cash := Cash + (J * Company[N].SharePrice)
                  end
              end;
            CloseWindow;
            CloseWindow      { TradeTitle }
          end;
    '3' : begin
            R := 0;              
            if (not BidUsed) and (random < R2) and (random < BidChance) then
              begin              {random < Percentage owned}
                R := random * R2 * MaxSharesBidded;
                Company[N].MaxStock := Company[N].MaxStock + trunc(R)
              end;
            BidUsed := true;
            TradeTitle;
            OpenWindow(2,19,78,5, Yellow,Blue,White, DoubleLineBorder);
            if trunc(R) = 0 then
              CentreText(Company[N].Name+' has refused to issue more shares.',1)
            else
              begin
                GotoXY((50-length(Company[N].Name)) div 2,1);
                write(Company[N].Name,' has issued ');
                TextColor(White);
                write(trunc(R));
                TextColor(Yellow);
                write(' more shares.')
              end;
            Continue;
            CloseWindow;
            CloseWindow         { Status }
          end
  end
end;

procedure DisplayStockExchange;

var
  I : integer;

begin
  OpenWindow(1,6,80,15, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Interstellar Stock Exchange  ',1);
  TextBackground(Blue);
  GotoXY((WinWidth-length(Player[CurrentPlayer].Name)-8) div 2,2);
  write('Player: ');
  TextColor(White);
  write(Player[CurrentPlayer].Name);
  TextColor(Yellow);
  writeln('.');
  AllOff := true;
  for I := 1 to MaxCompanies do
    if Company[I].OnMap then
      AllOff := false;
  if AllOff then
    CentreText('No companies on the map.',8)
  else
    begin
      GotoXY(1,4);
      TextColor(White);
      TextBackground(Black);
      ClrEOL;
      writeln('  Company                 Shares issued  Shares left  Price ($)  Return (%)');
      for I := 1 to MaxCompanies do
        if Company[I].OnMap then
          with Company[I] do
            begin
              TextColor(White);
              TextBackground(Red);
              write(Name[1]);
              SetDefaultColours;
              write(' ',Name);
              GotoXY(26,WhereY);
              write(StockIssued:10,'     ',MaxStock-StockIssued:10,'  ');
              writeln(SharePrice:11:2,'   ',Return*100:7:2)
            end
    end
end;

var
  OK : boolean;

begin { StockExchange }
  if QuitSelected or (not Player[CurrentPlayer].InGame) then
    exit;

  BidUsed := false;
  AdjustValues;
  if TotalValue(CurrentPlayer) <= -MaxOverdraft then
    begin
      OpenWindow(13,11,55,8, White,Red,Yellow, DoubleLineBorder);
      TextBackground(Black);
      CentreText('  Interstellar Trading Bank  ',1);
      TextBackground(Red);
      GotoXY(1,3);
      write('Your debt has amounted to ');
      TextColor(Yellow);
      write('$',Player[CurrentPlayer].Debt:4:2);
      TextColor(White);
      writeln('.');
      GotoXY(1,4);
      write('The bank has impounded ');
      TextColor(Yellow);
      write('$',Player[CurrentPlayer].Cash:4:2);
      TextColor(White);
      writeln(' from your cash.');
      Continue;
      CloseWindow;
      with Player[CurrentPlayer] do
        begin
          Debt := Debt / (InterestRate + 1) - Cash;
          Cash := 0.00;
          if Debt < 0.00 then
            Debt := 0.00
        end;
      if (TotalValue(CurrentPlayer) <= 0.00) and (random < ProbBankruptcy) then
        begin
          OpenWindow(19,11,43,8, White,Red,Yellow, DoubleLineBorder);
          TextBackground(Black);
          CentreText('  Interstellar Trading Bank  ',1);
          TextBackground(Red);
          CentreText('You have been declared bankrupt by the',3);
          CentreText('Interstellar Trading Bank.',4);
          Continue;
          DisplayStatus;
          CloseWindow;
          Player[CurrentPlayer].InGame := false;
          for I := 1 to MaxCompanies do
            with Company[I] do
              begin
                MaxStock := MaxStock + Player[CurrentPlayer].StockOwned[I];
                StockIssued := StockIssued - Player[CurrentPlayer].StockOwned[I];
                Player[CurrentPlayer].StockOwned[I] := 0
              end;
          Player[CurrentPlayer].Cash := 0.00;
          Player[CurrentPlayer].Debt := 0.00;
          AllOff := true;
          for I := 1 to NumberPlayers do
            if Player[I].InGame then
              AllOff := false;
          if AllOff then
            QuitSelected := true
        end
    end;

  if QuitSelected or (not Player[CurrentPlayer].InGame) then
    exit;

  DisplayStockExchange;
  repeat
    OpenWindow(1,21,80,5, Yellow,Blue,White, DoubleLineBorder);
    GotoXY(1,2);
    TextBackground(Black);
    write('<1>');
    TextBackground(Blue);
    write(' Display stock owned.');
    GotoXY(38,2);
    TextBackground(Black);
    write('<2>');
    TextBackground(Blue);
    write(' Display map.');
    GotoXY(1,3);
    TextBackground(Black);
    write('<3>');
    TextBackground(Blue);
    write(' Visit the Trading Bank.');
    GotoXY(38,3);
    TextBackground(Black);
    write('<4>');
    TextBackground(Blue);
    write(' Exit the Stock Exchange.');
    GotoXY(18,1);
    ClrEOL;
    write('Enter selection [');
    TextColor(White);
    write('Company letter');
    TextColor(Yellow);
    write('/');
    TextBackground(Black);
    write('1');
    TextBackground(Blue);
    write('-');
    TextBackground(Black);
    write('4');
    TextBackground(Blue);
    write(']: ');
    SetCursorOn(true);
    repeat
      C := UKey;
      if C in ['1'..'4',' '] then
        OK := true
      else
        begin
          OK := false;
          if (C >= 'A') and (C <= chr(MaxCompanies+$40))
             and Company[ord(C)-$40].OnMap then
            OK := true
        end
    until OK;
    SetCursorOn(false);
    CloseWindow; { Select option }
    case C of
      '1'      : DisplayStatus;
      '2'      : begin
                   DisplayMap;
                   OpenWindow(1,20,80,6,LightGray,Black,LightGray, NoBorder);
                   write('');
                   for I := 1 to 78 do
                     write('');
                   write('');
                   GotoXY(80,WhereY);
                   write('');
                   for I := 1 to 78 do
                     write('');
                   write('');
                   CentreText('[ Press <SPACE> to continue ]',2);
                   WaitForKey;
                   CloseWindow;
                   CloseWindow  { Map }
                 end;
      '3'      : VisitBank;
      '4', ' ' : C := ' '
    else
      Trade(ord(C)-$40);
      CloseWindow;      { Stock exchange }
      DisplayStockExchange
    end
  until C = ' ';
  CloseWindow  { Stock exchange }
end;

procedure DisplayStatusOfAll;    { Display everyone's status at end of game }

var
  I, J : integer;
  C    : char;

procedure SortPlayers;             { Sort players according to total value }

var
  I, J, K : integer;
  Temp    : PlayerRec;
  R1, R2  : float;

begin
  for I := 1 to NumberPlayers - 1 do
    begin
      K := I;
      Temp := Player[I];
      for J := I+1 to NumberPlayers do
        if TotalValue(K) < TotalValue(J) then
          begin
            K := J;
            Temp := Player[J]
          end;
      Player[K] := Player[I];
      Player[I] := Temp
    end
end;

begin { DisplayStatus_of_all }
  OpenWindow(21,11,37,7, White,Red,Yellow, DoubleLineBorder);
  TextBackground(Black);
  CentreText('  Game Over  ',1);
  TextBackground(Red);
  GotoXY(1,3);
  write('The game is over after ');
  TextColor(Yellow);
  write(TurnNumber-1);
  TextColor(White);
  write(' turns.');
  Continue;
  CloseWindow;
  for I := 1 to NumberPlayers do
    begin
      CurrentPlayer := I;
      DisplayStatus
    end;
  OpenWindow(6,7,70,NumberPlayers+10, Yellow,Blue,White, DoubleLineBorder);
  TextBackground(Black);
  if NumberPlayers = 1 then
    begin
      CentreText('  Total value  ',1);
      TextBackground(Blue);
      GotoXY(17,5);
      write('Your total value was ');
      TextBackground(Black);
      write(' $',TotalValue(1):4:2,' ');
      TextBackground(Blue)
    end
  else
    begin
      CentreText('  Game Winner  ',1);
      TextBackground(Blue);
      GotoXY(1,3);
      SortPlayers;
      GotoXY((52-length(Player[1].Name)) div 2,3);
      write('The winner is ');
      TextColor(White);
      write(Player[1].Name);
      TextColor(Yellow);
      write(',');
      if TotalValue(1) = 0.00 then
        begin
          GotoXY(20,4);
          write('who is ');
          TextColor(Yellow+Blink);
          TextBackground(Red);
          write(' *** Bankrupt *** ')
        end
      else
        begin
          GotoXY(20,4);
          write('with a value of ');
          TextBackground(Black);
          write(' $',TotalValue(1):4:2,' ')
        end;
      TextColor(White);
      TextBackground(Black);
      GotoXY(1,6);
      ClrEOL;
      writeln(' Position Player');
      GotoXY(50,6);
      writeln('Total value ($)');
      SetDefaultColours;
      for I := 1 to NumberPlayers do
        begin
          TextColor(White);
          write('    ',I);
          case I of
            1 : write('st   ');
            2 : write('nd   ');
            3 : write('rd   ')
          else
            write('th   ')
          end;
          TextColor(Yellow);
          write(Player[I].Name);
          GotoXY(48,WhereY);
          if TotalValue(I) = 0.00 then
            begin
              TextColor(Yellow);
              TextBackground(Red);
              writeln(' *** BANKRUPT *** ');
              SetDefaultColours
            end
          else
            writeln(TotalValue(I):15:2)
        end
    end;
  Continue;
  CloseWindow
end;


begin  { Main }
  Init;
  InitPlayers;
  InitData;
  repeat
    SelectMoves;
    GetMove;
    ProcessMove;
    StockExchange;
    NextPlayer
  until (TurnNumber = MaxTurn + 1) or QuitSelected;
  DisplayStatusOfAll;
  DeInit
end.
